# Copyright 2012-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Parsing for router status entries, the information for individual routers
within a network status document. This information is provided from a few
sources...

* control port via 'GETINFO ns/\\*' and 'GETINFO md/\\*' queries
* router entries in a network status document, like the cached-consensus

**Module Overview:**

::

  RouterStatusEntry - Common parent for router status entries
    |- RouterStatusEntryV2 - Entry for a network status v2 document
    |   +- RouterStatusEntryBridgeV2 - Entry for a bridge flavored v2 document
    |
    |- RouterStatusEntryV3 - Entry for a network status v3 document
    +- RouterStatusEntryMicroV3 - Entry for a microdescriptor flavored v3 document
"""

import binascii
import io

import stem.exit_policy
import stem.prereq
import stem.util.str_tools

from stem.descriptor import (
  KEYWORD_LINE,
  Descriptor,
  _descriptor_content,
  _value,
  _values,
  _descriptor_components,
  _parse_protocol_line,
  _read_until_keywords,
  _random_nickname,
  _random_ipv4_address,
  _random_date,
)

_parse_pr_line = _parse_protocol_line('pr', 'protocols')


def _parse_file(document_file, validate, entry_class, entry_keyword = 'r', start_position = None, end_position = None, section_end_keywords = (), extra_args = ()):
  """
  Reads a range of the document_file containing some number of entry_class
  instances. We deliminate the entry_class entries by the keyword on their
  first line (entry_keyword). When finished the document is left at the
  end_position.

  Either an end_position or section_end_keywords must be provided.

  :param file document_file: file with network status document content
  :param bool validate: checks the validity of the document's contents if
    **True**, skips these checks otherwise
  :param class entry_class: class to construct instance for
  :param str entry_keyword: first keyword for the entry instances
  :param int start_position: start of the section, default is the current position
  :param int end_position: end of the section
  :param tuple section_end_keywords: keyword(s) that deliminate the end of the
    section if no end_position was provided
  :param tuple extra_args: extra arguments for the entry_class (after the
    content and validate flag)

  :returns: iterator over entry_class instances

  :raises:
    * **ValueError** if the contents is malformed and validate is **True**
    * **IOError** if the file can't be read
  """

  if start_position:
    document_file.seek(start_position)
  else:
    start_position = document_file.tell()

  # check if we're starting at the end of the section (ie, there's no entries to read)
  if section_end_keywords:
    first_keyword = None
    line_match = KEYWORD_LINE.match(stem.util.str_tools._to_unicode(document_file.readline()))

    if line_match:
      first_keyword = line_match.groups()[0]

    document_file.seek(start_position)

    if first_keyword in section_end_keywords:
      return

  while end_position is None or document_file.tell() < end_position:
    desc_lines, ending_keyword = _read_until_keywords(
      (entry_keyword,) + section_end_keywords,
      document_file,
      ignore_first = True,
      end_position = end_position,
      include_ending_keyword = True
    )

    desc_content = bytes.join(b'', desc_lines)

    if desc_content:
      yield entry_class(desc_content, validate, *extra_args)

      # check if we stopped at the end of the section
      if ending_keyword in section_end_keywords:
        break
    else:
      break


def _parse_r_line(descriptor, entries):
  # Parses a RouterStatusEntry's 'r' line. They're very nearly identical for
  # all current entry types (v2, v3, and microdescriptor v3) with one little
  # wrinkle: only the microdescriptor flavor excludes a 'digest' field.
  #
  # For v2 and v3 router status entries:
  #   "r" nickname identity digest publication IP ORPort DirPort
  #   example: r mauer BD7xbfsCFku3+tgybEZsg8Yjhvw itcuKQ6PuPLJ7m/Oi928WjO2j8g 2012-06-22 13:19:32 80.101.105.103 9001 0
  #
  # For v3 microdescriptor router status entries:
  #   "r" nickname identity publication IP ORPort DirPort
  #   example: r Konata ARIJF2zbqirB9IwsW0mQznccWww 2012-09-24 13:40:40 69.64.48.168 9001 9030

  value = _value('r', entries)
  include_digest = not isinstance(descriptor, RouterStatusEntryMicroV3)

  r_comp = value.split(' ')

  # inject a None for the digest to normalize the field positioning
  if not include_digest:
    r_comp.insert(2, None)

  if len(r_comp) < 8:
    expected_field_count = 'eight' if include_digest else 'seven'
    raise ValueError("%s 'r' line must have %s values: r %s" % (descriptor._name(), expected_field_count, value))

  if not stem.util.tor_tools.is_valid_nickname(r_comp[0]):
    raise ValueError("%s nickname isn't valid: %s" % (descriptor._name(), r_comp[0]))
  elif not stem.util.connection.is_valid_ipv4_address(r_comp[5]):
    raise ValueError("%s address isn't a valid IPv4 address: %s" % (descriptor._name(), r_comp[5]))
  elif not stem.util.connection.is_valid_port(r_comp[6]):
    raise ValueError('%s ORPort is invalid: %s' % (descriptor._name(), r_comp[6]))
  elif not stem.util.connection.is_valid_port(r_comp[7], allow_zero = True):
    raise ValueError('%s DirPort is invalid: %s' % (descriptor._name(), r_comp[7]))

  descriptor.nickname = r_comp[0]
  descriptor.fingerprint = _base64_to_hex(r_comp[1])

  if include_digest:
    descriptor.digest = _base64_to_hex(r_comp[2])

  descriptor.address = r_comp[5]
  descriptor.or_port = int(r_comp[6])
  descriptor.dir_port = None if r_comp[7] == '0' else int(r_comp[7])

  try:
    published = '%s %s' % (r_comp[3], r_comp[4])
    descriptor.published = stem.util.str_tools._parse_timestamp(published)
  except ValueError:
    raise ValueError("Publication time time wasn't parsable: r %s" % value)


def _parse_a_line(descriptor, entries):
  # "a" SP address ":" portlist
  # example: a [2001:888:2133:0:82:94:251:204]:9001

  or_addresses = []

  for value in _values('a', entries):
    if ':' not in value:
      raise ValueError("%s 'a' line must be of the form '[address]:[ports]': a %s" % (descriptor._name(), value))

    address, port = value.rsplit(':', 1)

    if not stem.util.connection.is_valid_ipv4_address(address) and not stem.util.connection.is_valid_ipv6_address(address, allow_brackets = True):
      raise ValueError("%s 'a' line must start with an IPv6 address: a %s" % (descriptor._name(), value))

    if stem.util.connection.is_valid_port(port):
      or_addresses.append((address.lstrip('[').rstrip(']'), int(port), stem.util.connection.is_valid_ipv6_address(address, allow_brackets = True)))
    else:
      raise ValueError("%s 'a' line had an invalid port (%s): a %s" % (descriptor._name(), port, value))

  descriptor.or_addresses = or_addresses


def _parse_s_line(descriptor, entries):
  # "s" Flags
  # example: s Named Running Stable Valid

  value = _value('s', entries)
  flags = [] if value == '' else value.split(' ')
  descriptor.flags = flags

  for flag in flags:
    if flags.count(flag) > 1:
      raise ValueError('%s had duplicate flags: s %s' % (descriptor._name(), value))
    elif flag == '':
      raise ValueError("%s had extra whitespace on its 's' line: s %s" % (descriptor._name(), value))


def _parse_v_line(descriptor, entries):
  # "v" version
  # example: v Tor 0.2.2.35
  #
  # The spec says that if this starts with "Tor " then what follows is a
  # tor version. If not then it has "upgraded to a more sophisticated
  # protocol versioning system".

  value = _value('v', entries)
  descriptor.version_line = value

  if value.startswith('Tor '):
    try:
      descriptor.version = stem.version._get_version(value[4:])
    except ValueError as exc:
      raise ValueError('%s has a malformed tor version (%s): v %s' % (descriptor._name(), exc, value))


def _parse_w_line(descriptor, entries):
  # "w" "Bandwidth=" INT ["Measured=" INT] ["Unmeasured=1"]
  # example: w Bandwidth=7980

  value = _value('w', entries)
  w_comp = value.split(' ')

  if len(w_comp) < 1:
    raise ValueError("%s 'w' line is blank: w %s" % (descriptor._name(), value))
  elif not w_comp[0].startswith('Bandwidth='):
    raise ValueError("%s 'w' line needs to start with a 'Bandwidth=' entry: w %s" % (descriptor._name(), value))

  bandwidth = None
  measured = None
  is_unmeasured = False
  unrecognized_bandwidth_entries = []

  for w_entry in w_comp:
    if '=' in w_entry:
      w_key, w_value = w_entry.split('=', 1)
    else:
      w_key, w_value = w_entry, None

    if w_key == 'Bandwidth':
      if not (w_value and w_value.isdigit()):
        raise ValueError("%s 'Bandwidth=' entry needs to have a numeric value: w %s" % (descriptor._name(), value))

      bandwidth = int(w_value)
    elif w_key == 'Measured':
      if not (w_value and w_value.isdigit()):
        raise ValueError("%s 'Measured=' entry needs to have a numeric value: w %s" % (descriptor._name(), value))

      measured = int(w_value)
    elif w_key == 'Unmeasured':
      if w_value != '1':
        raise ValueError("%s 'Unmeasured=' should only have the value of '1': w %s" % (descriptor._name(), value))

      is_unmeasured = True
    else:
      unrecognized_bandwidth_entries.append(w_entry)

  descriptor.bandwidth = bandwidth
  descriptor.measured = measured
  descriptor.is_unmeasured = is_unmeasured
  descriptor.unrecognized_bandwidth_entries = unrecognized_bandwidth_entries


def _parse_p_line(descriptor, entries):
  # "p" ("accept" / "reject") PortList
  #
  # examples:
  #
  #   p accept 80,110,143,443,993,995,6660-6669,6697,7000-7001
  #   p reject 1-65535

  value = _value('p', entries)

  try:
    descriptor.exit_policy = stem.exit_policy.MicroExitPolicy(value)
  except ValueError as exc:
    raise ValueError('%s exit policy is malformed (%s): p %s' % (descriptor._name(), exc, value))


def _parse_id_line(descriptor, entries):
  # "id" "ed25519" ed25519-identity
  #
  # examples:
  #
  #   id ed25519 none
  #   id ed25519 8RH34kO07Pp+XYwzdoATVyCibIvmbslUjRkAm7J4IA8

  value = _value('id', entries)

  if value:
    if descriptor.document and not descriptor.document.is_vote:
      raise ValueError("%s 'id' line should only appear in votes: id %s" % (descriptor._name(), value))

    value_comp = value.split()

    if len(value_comp) >= 2:
      descriptor.identifier_type = value_comp[0]
      descriptor.identifier = value_comp[1]
    else:
      raise ValueError("'id' lines should contain both the key type and digest: id %s" % value)


def _parse_m_line(descriptor, entries):
  # "m" methods 1*(algorithm "=" digest)
  # example: m 8,9,10,11,12 sha256=g1vx9si329muxV3tquWIXXySNOIwRGMeAESKs/v4DWs

  all_hashes = []

  for value in _values('m', entries):
    m_comp = value.split(' ')

    if not (descriptor.document and descriptor.document.is_vote):
      vote_status = 'vote' if descriptor.document else '<undefined document>'
      raise ValueError("%s 'm' line should only appear in votes (appeared in a %s): m %s" % (descriptor._name(), vote_status, value))
    elif len(m_comp) < 1:
      raise ValueError("%s 'm' line needs to start with a series of methods: m %s" % (descriptor._name(), value))

    try:
      methods = [int(entry) for entry in m_comp[0].split(',')]
    except ValueError:
      raise ValueError('%s microdescriptor methods should be a series of comma separated integers: m %s' % (descriptor._name(), value))

    hashes = {}

    for entry in m_comp[1:]:
      if '=' not in entry:
        raise ValueError("%s can only have a series of 'algorithm=digest' mappings after the methods: m %s" % (descriptor._name(), value))

      hash_name, digest = entry.split('=', 1)
      hashes[hash_name] = digest

    all_hashes.append((methods, hashes))

  descriptor.microdescriptor_hashes = all_hashes


def _parse_microdescriptor_m_line(descriptor, entries):
  # "m" digest
  # example: m aiUklwBrua82obG5AsTX+iEpkjQA2+AQHxZ7GwMfY70

  descriptor.microdescriptor_digest = _value('m', entries)

  # TODO: drop the following in stem 2.x

  descriptor.digest = _base64_to_hex(_value('m', entries), check_if_fingerprint = False)


def _base64_to_hex(identity, check_if_fingerprint = True):
  """
  Decodes a base64 value to hex. For example...

  ::

    >>> _base64_to_hex('p1aag7VwarGxqctS7/fS0y5FU+s')
    'A7569A83B5706AB1B1A9CB52EFF7D2D32E4553EB'

  :param str identity: encoded fingerprint from the consensus
  :param bool check_if_fingerprint: asserts that the result is a fingerprint if **True**

  :returns: **str** with the uppercase hex encoding of the relay's fingerprint

  :raises: **ValueError** if the result isn't a valid fingerprint
  """

  try:
    identity_decoded = stem.util.str_tools._decode_b64(stem.util.str_tools._to_bytes(identity))
  except (TypeError, binascii.Error):
    raise ValueError("Unable to decode identity string '%s'" % identity)

  fingerprint = binascii.hexlify(identity_decoded).upper()

  if stem.prereq.is_python_3():
    fingerprint = stem.util.str_tools._to_unicode(fingerprint)

  if check_if_fingerprint:
    if not stem.util.tor_tools.is_valid_fingerprint(fingerprint):
      raise ValueError("Decoded '%s' to be '%s', which isn't a valid fingerprint" % (identity, fingerprint))

  return fingerprint


class RouterStatusEntry(Descriptor):
  """
  Information about an individual router stored within a network status
  document. This is the common parent for concrete status entry types.

  :var stem.descriptor.networkstatus.NetworkStatusDocument document: **\\*** document that this descriptor came from

  :var str nickname: **\\*** router's nickname
  :var str fingerprint: **\\*** router's fingerprint
  :var datetime published: **\\*** router's publication
  :var str address: **\\*** router's IP address
  :var int or_port: **\\*** router's ORPort
  :var int dir_port: **\\*** router's DirPort

  :var list flags: **\\*** list of :data:`~stem.Flag` associated with the relay

  :var stem.version.Version version: parsed version of tor, this is **None** if
    the relay's using a new versioning scheme
  :var str version_line: versioning information reported by the relay
  """

  ATTRIBUTES = {
    'nickname': (None, _parse_r_line),
    'fingerprint': (None, _parse_r_line),
    'published': (None, _parse_r_line),
    'address': (None, _parse_r_line),
    'or_port': (None, _parse_r_line),
    'dir_port': (None, _parse_r_line),

    'flags': (None, _parse_s_line),

    'version_line': (None, _parse_v_line),
    'version': (None, _parse_v_line),
  }

  PARSER_FOR_LINE = {
    'r': _parse_r_line,
    's': _parse_s_line,
    'v': _parse_v_line,
  }

  @classmethod
  def from_str(cls, content, **kwargs):
    # Router status entries don't have their own @type annotation, so to make
    # our subclass from_str() work we need to do the type inferencing ourself.

    if cls == RouterStatusEntry:
      raise NotImplementedError('Please use the from_str() method from RouterStatusEntry subclasses, not RouterStatusEntry itself')
    elif 'descriptor_type' in kwargs:
      raise ValueError("Router status entries don't have their own @type annotation. As such providing a 'descriptor_type' argument with RouterStatusEntry.from_str() does not work. Please drop the 'descriptor_type' argument when using this these subclasses' from_str() method.")

    is_multiple = kwargs.pop('multiple', False)
    validate = kwargs.pop('validate', False)
    results = list(_parse_file(io.BytesIO(stem.util.str_tools._to_bytes(content)), validate, cls, **kwargs))

    if is_multiple:
      return results
    elif len(results) == 1:
      return results[0]
    else:
      raise ValueError("Descriptor.from_str() expected a single descriptor, but had %i instead. Please include 'multiple = True' if you want a list of results instead." % len(results))

  def __init__(self, content, validate = False, document = None):
    """
    Parse a router descriptor in a network status document.

    :param str content: router descriptor content to be parsed
    :param NetworkStatusDocument document: document this descriptor came from
    :param bool validate: checks the validity of the content if **True**, skips
      these checks otherwise

    :raises: **ValueError** if the descriptor data is invalid
    """

    super(RouterStatusEntry, self).__init__(content, lazy_load = not validate)
    self.document = document
    entries = _descriptor_components(content, validate)

    if validate:
      for keyword in self._required_fields():
        if keyword not in entries:
          raise ValueError("%s must have a '%s' line:\n%s" % (self._name(True), keyword, str(self)))

      for keyword in self._single_fields():
        if keyword in entries and len(entries[keyword]) > 1:
          raise ValueError("%s can only have a single '%s' line, got %i:\n%s" % (self._name(True), keyword, len(entries[keyword]), str(self)))

      if 'r' != list(entries.keys())[0]:
        raise ValueError("%s are expected to start with a 'r' line:\n%s" % (self._name(True), str(self)))

      self._parse(entries, validate)
    else:
      self._entries = entries

  def _name(self, is_plural = False):
    """
    Name for this descriptor type.
    """

    return 'Router status entries' if is_plural else 'Router status entry'

  def _required_fields(self):
    """
    Provides lines that must appear in the descriptor.
    """

    return ()

  def _single_fields(self):
    """
    Provides lines that can only appear in the descriptor once.
    """

    return ()


class RouterStatusEntryV2(RouterStatusEntry):
  """
  Information about an individual router stored within a version 2 network
  status document.

  :var str digest: **\\*** router's upper-case hex digest

  **\\*** attribute is either required when we're parsed with validation or has
  a default value, others are left as **None** if undefined
  """

  TYPE_ANNOTATION_NAME = 'network-status-consensus-2'

  ATTRIBUTES = dict(RouterStatusEntry.ATTRIBUTES, **{
    'digest': (None, _parse_r_line),
  })

  @classmethod
  def content(cls, attr = None, exclude = (), sign = False):
    if sign:
      raise NotImplementedError('Signing of %s not implemented' % cls.__name__)

    return _descriptor_content(attr, exclude, (
      ('r', '%s p1aag7VwarGxqctS7/fS0y5FU+s oQZFLYe9e4A7bOkWKR7TaNxb0JE %s %s 9001 0' % (_random_nickname(), _random_date(), _random_ipv4_address())),
    ))

  def _name(self, is_plural = False):
    return 'Router status entries (v2)' if is_plural else 'Router status entry (v2)'

  def _required_fields(self):
    return ('r')

  def _single_fields(self):
    return ('r', 's', 'v')


class RouterStatusEntryBridgeV2(RouterStatusEntryV2):
  """
  Information about an individual router stored within a bridge flavored
  version 2 network status document.

  .. versionadded:: 1.8.0
  """

  TYPE_ANNOTATION_NAME = 'bridge-network-status'


class RouterStatusEntryV3(RouterStatusEntry):
  """
  Information about an individual router stored within a version 3 network
  status document.

  :var list or_addresses: **\\*** relay's OR addresses, this is a tuple listing
    of the form (address (**str**), port (**int**), is_ipv6 (**bool**))
  :var str identifier_type: identity digest key type
  :var str identifier: base64 encoded identity digest
  :var str digest: **\\*** router's upper-case hex digest

  :var int bandwidth: bandwidth measured to be available by the relay, this is
    an arbitrary units (currently kilobytes per second) heuristic generated by
    the Bandwidth authoritites to weight relay selection
  :var int measured: *bandwidth* vote provided by a bandwidth authority
  :var bool is_unmeasured: *bandwidth* measurement isn't based on three or more
    measurements
  :var list unrecognized_bandwidth_entries: **\\*** bandwidth weighting
    information that isn't yet recognized

  :var stem.exit_policy.MicroExitPolicy exit_policy: router's exit policy
  :var dict protocols: mapping of protocols to their supported versions

  :var list microdescriptor_hashes: **\\*** tuples of two values, the list of
    consensus methods for generating a set of digests and the 'algorithm =>
    digest' mappings

  **\\*** attribute is either required when we're parsed with validation or has
  a default value, others are left as **None** if undefined

  .. versionchanged:: 1.5.0
     Added the identifier and identifier_type attributes.

  .. versionchanged:: 1.6.0
     Added the protocols attribute.
  """

  TYPE_ANNOTATION_NAME = 'network-status-consensus-3'

  ATTRIBUTES = dict(RouterStatusEntry.ATTRIBUTES, **{
    'digest': (None, _parse_r_line),
    'or_addresses': ([], _parse_a_line),
    'identifier_type': (None, _parse_id_line),
    'identifier': (None, _parse_id_line),

    'bandwidth': (None, _parse_w_line),
    'measured': (None, _parse_w_line),
    'is_unmeasured': (False, _parse_w_line),
    'unrecognized_bandwidth_entries': ([], _parse_w_line),

    'exit_policy': (None, _parse_p_line),
    'protocols': ({}, _parse_pr_line),
    'microdescriptor_hashes': ([], _parse_m_line),
  })

  PARSER_FOR_LINE = dict(RouterStatusEntry.PARSER_FOR_LINE, **{
    'a': _parse_a_line,
    'w': _parse_w_line,
    'p': _parse_p_line,
    'pr': _parse_pr_line,
    'id': _parse_id_line,
    'm': _parse_m_line,
  })

  @classmethod
  def content(cls, attr = None, exclude = (), sign = False):
    if sign:
      raise NotImplementedError('Signing of %s not implemented' % cls.__name__)

    return _descriptor_content(attr, exclude, (
      ('r', '%s p1aag7VwarGxqctS7/fS0y5FU+s oQZFLYe9e4A7bOkWKR7TaNxb0JE %s %s 9001 0' % (_random_nickname(), _random_date(), _random_ipv4_address())),
      ('s', 'Fast Named Running Stable Valid'),
    ))

  def _name(self, is_plural = False):
    return 'Router status entries (v3)' if is_plural else 'Router status entry (v3)'

  def _required_fields(self):
    return ('r', 's')

  def _single_fields(self):
    return ('r', 's', 'v', 'w', 'p', 'pr')


class RouterStatusEntryMicroV3(RouterStatusEntry):
  """
  Information about an individual router stored within a microdescriptor
  flavored network status document.

  :var list or_addresses: **\\*** relay's OR addresses, this is a tuple listing
    of the form (address (**str**), port (**int**), is_ipv6 (**bool**))
  :var int bandwidth: bandwidth claimed by the relay (in kb/s)
  :var int measured: bandwidth measured to be available by the relay
  :var bool is_unmeasured: bandwidth measurement isn't based on three or more
    measurements
  :var list unrecognized_bandwidth_entries: **\\*** bandwidth weighting
    information that isn't yet recognized
  :var dict protocols: mapping of protocols to their supported versions

  :var str digest: **\\*** router's hex encoded digest of our corresponding
    microdescriptor (**deprecated**, use microdescriptor_digest instead)
  :var str microdescriptor_digest: **\\*** router's base64 encoded digest of our corresponding microdescriptor

  .. versionchanged:: 1.6.0
     Added the protocols attribute.

  .. versionchanged:: 1.7.0
     Added the or_addresses attribute.

  .. versionchanged:: 1.7.0
     Added the microdescriptor_digest attribute to replace our now deprecated digest attribute.

  **\\*** attribute is either required when we're parsed with validation or has
  a default value, others are left as **None** if undefined
  """

  TYPE_ANNOTATION_NAME = 'network-status-microdesc-consensus-3'

  ATTRIBUTES = dict(RouterStatusEntry.ATTRIBUTES, **{
    'or_addresses': ([], _parse_a_line),
    'bandwidth': (None, _parse_w_line),
    'measured': (None, _parse_w_line),
    'is_unmeasured': (False, _parse_w_line),
    'unrecognized_bandwidth_entries': ([], _parse_w_line),
    'protocols': ({}, _parse_pr_line),

    'microdescriptor_digest': (None, _parse_microdescriptor_m_line),
    'digest': (None, _parse_microdescriptor_m_line),
  })

  PARSER_FOR_LINE = dict(RouterStatusEntry.PARSER_FOR_LINE, **{
    'a': _parse_a_line,
    'w': _parse_w_line,
    'm': _parse_microdescriptor_m_line,
    'pr': _parse_pr_line,
  })

  @classmethod
  def content(cls, attr = None, exclude = (), sign = False):
    if sign:
      raise NotImplementedError('Signing of %s not implemented' % cls.__name__)

    return _descriptor_content(attr, exclude, (
      ('r', '%s ARIJF2zbqirB9IwsW0mQznccWww %s %s 9001 9030' % (_random_nickname(), _random_date(), _random_ipv4_address())),
      ('m', 'aiUklwBrua82obG5AsTX+iEpkjQA2+AQHxZ7GwMfY70'),
      ('s', 'Fast Guard HSDir Named Running Stable V2Dir Valid'),
    ))

  def _name(self, is_plural = False):
    return 'Router status entries (micro v3)' if is_plural else 'Router status entry (micro v3)'

  def _required_fields(self):
    return ('r', 's', 'm')

  def _single_fields(self):
    return ('r', 's', 'v', 'w', 'm', 'pr')



# Copyright 2013-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Parsing for Tor microdescriptors, which contain a distilled version of a
relay's server descriptor. As of Tor version 0.2.3.3-alpha Tor no longer
downloads server descriptors by default, opting for microdescriptors instead.

Unlike most descriptor documents these aren't available on the metrics site
(since they don't contain any information that the server descriptors don't).

The limited information in microdescriptors make them rather clunky to use
compared with server descriptors. For instance microdescriptors lack the
relay's fingerprint, making it difficut to use them to look up the relay's
other descriptors.

To do so you need to match the microdescriptor's digest against its
corresponding router status entry. For added fun as of this writing the
controller doesn't even surface those router status entries
(:trac:`7953`).

For instance, here's an example that prints the nickname and fingerprints of
the exit relays.

::

  import os

  from stem.control import Controller
  from stem.descriptor import parse_file

  with Controller.from_port(port = 9051) as controller:
    controller.authenticate()

    exit_digests = set()
    data_dir = controller.get_conf('DataDirectory')

    for desc in controller.get_microdescriptors():
      if desc.exit_policy.is_exiting_allowed():
        exit_digests.add(desc.digest)

    print 'Exit Relays:'

    for desc in parse_file(os.path.join(data_dir, 'cached-microdesc-consensus')):
      if desc.digest in exit_digests:
        print '  %s (%s)' % (desc.nickname, desc.fingerprint)

Doing the same is trivial with server descriptors...

::

  from stem.descriptor import parse_file

  print 'Exit Relays:'

  for desc in parse_file('/home/atagar/.tor/cached-descriptors'):
    if desc.exit_policy.is_exiting_allowed():
      print '  %s (%s)' % (desc.nickname, desc.fingerprint)

**Module Overview:**

::

  Microdescriptor - Tor microdescriptor.
"""

import hashlib

import stem.exit_policy
import stem.prereq

from stem.descriptor import (
  Descriptor,
  DigestHash,
  DigestEncoding,
  _descriptor_content,
  _descriptor_components,
  _read_until_keywords,
  _values,
  _parse_simple_line,
  _parse_protocol_line,
  _parse_key_block,
  _random_crypto_blob,
)

from stem.descriptor.router_status_entry import (
  _parse_a_line,
  _parse_p_line,
)

if stem.prereq._is_lru_cache_available():
  from functools import lru_cache
else:
  from stem.util.lru_cache import lru_cache

REQUIRED_FIELDS = (
  'onion-key',
)

SINGLE_FIELDS = (
  'onion-key',
  'ntor-onion-key',
  'family',
  'p',
  'p6',
  'pr',
)


def _parse_file(descriptor_file, validate = False, **kwargs):
  """
  Iterates over the microdescriptors in a file.

  :param file descriptor_file: file with descriptor content
  :param bool validate: checks the validity of the descriptor's content if
    **True**, skips these checks otherwise
  :param dict kwargs: additional arguments for the descriptor constructor

  :returns: iterator for Microdescriptor instances in the file

  :raises:
    * **ValueError** if the contents is malformed and validate is True
    * **IOError** if the file can't be read
  """

  while True:
    annotations = _read_until_keywords('onion-key', descriptor_file)

    # read until we reach an annotation or onion-key line
    descriptor_lines = []

    # read the onion-key line, done if we're at the end of the document

    onion_key_line = descriptor_file.readline()

    if onion_key_line:
      descriptor_lines.append(onion_key_line)
    else:
      break

    while True:
      last_position = descriptor_file.tell()
      line = descriptor_file.readline()

      if not line:
        break  # EOF
      elif line.startswith(b'@') or line.startswith(b'onion-key'):
        descriptor_file.seek(last_position)
        break
      else:
        descriptor_lines.append(line)

    if descriptor_lines:
      if descriptor_lines[0].startswith(b'@type'):
        descriptor_lines = descriptor_lines[1:]

      # strip newlines from annotations
      annotations = list(map(bytes.strip, annotations))

      descriptor_text = bytes.join(b'', descriptor_lines)

      yield Microdescriptor(descriptor_text, validate, annotations, **kwargs)
    else:
      break  # done parsing descriptors


def _parse_id_line(descriptor, entries):
  identities = {}

  for entry in _values('id', entries):
    entry_comp = entry.split()

    if len(entry_comp) >= 2:
      key_type, key_value = entry_comp[0], entry_comp[1]

      if key_type in identities:
        raise ValueError("There can only be one 'id' line per a key type, but '%s' appeared multiple times" % key_type)

      descriptor.identifier_type = key_type
      descriptor.identifier = key_value
      identities[key_type] = key_value
    else:
      raise ValueError("'id' lines should contain both the key type and digest: id %s" % entry)

  descriptor.identifiers = identities


_parse_onion_key_line = _parse_key_block('onion-key', 'onion_key', 'RSA PUBLIC KEY')
_parse_ntor_onion_key_line = _parse_simple_line('ntor-onion-key', 'ntor_onion_key')
_parse_family_line = _parse_simple_line('family', 'family', func = lambda v: v.split(' '))
_parse_p6_line = _parse_simple_line('p6', 'exit_policy_v6', func = lambda v: stem.exit_policy.MicroExitPolicy(v))
_parse_pr_line = _parse_protocol_line('pr', 'protocols')


class Microdescriptor(Descriptor):
  """
  Microdescriptor (`descriptor specification
  <https://gitweb.torproject.org/torspec.git/tree/dir-spec.txt>`_)

  :var str onion_key: **\\*** key used to encrypt EXTEND cells
  :var str ntor_onion_key: base64 key used to encrypt EXTEND in the ntor protocol
  :var list or_addresses: **\\*** alternative for our address/or_port attributes, each
    entry is a tuple of the form (address (**str**), port (**int**), is_ipv6
    (**bool**))
  :var list family: **\\*** nicknames or fingerprints of declared family
  :var stem.exit_policy.MicroExitPolicy exit_policy: **\\*** relay's exit policy
  :var stem.exit_policy.MicroExitPolicy exit_policy_v6: **\\*** exit policy for IPv6
  :var hash identifiers: mapping of key types (like rsa1024 or ed25519) to
    their base64 encoded identity, this is only used for collision prevention
    (:trac:`11743`)
  :var dict protocols: mapping of protocols to their supported versions

  :var str identifier: base64 encoded identity digest (**deprecated**, use
    identifiers instead)
  :var str identifier_type: identity digest key type (**deprecated**, use
    identifiers instead)

  **\\*** attribute is required when we're parsed with validation

  .. versionchanged:: 1.1.0
     Added the identifier and identifier_type attributes.

  .. versionchanged:: 1.5.0
     Added the identifiers attribute, and deprecated identifier and
     identifier_type since the field can now appear multiple times.

  .. versionchanged:: 1.6.0
     Added the protocols attribute.

  .. versionchanged:: 1.8.0
     Replaced our **digest** attribute with a much more flexible **digest()**
     method. Unfortunately I cannot do this in a backward compatible way
     because of the name conflict. The old digest had multiple problems (for
     instance, being hex rather than base64 encoded), so hopefully no one was
     using it. Very sorry if this causes trouble for anyone.
  """

  TYPE_ANNOTATION_NAME = 'microdescriptor'

  ATTRIBUTES = {
    'onion_key': (None, _parse_onion_key_line),
    'ntor_onion_key': (None, _parse_ntor_onion_key_line),
    'or_addresses': ([], _parse_a_line),
    'family': ([], _parse_family_line),
    'exit_policy': (stem.exit_policy.MicroExitPolicy('reject 1-65535'), _parse_p_line),
    'exit_policy_v6': (None, _parse_p6_line),
    'identifier_type': (None, _parse_id_line),  # deprecated in favor of identifiers
    'identifier': (None, _parse_id_line),  # deprecated in favor of identifiers
    'identifiers': ({}, _parse_id_line),
    'protocols': ({}, _parse_pr_line),
  }

  PARSER_FOR_LINE = {
    'onion-key': _parse_onion_key_line,
    'ntor-onion-key': _parse_ntor_onion_key_line,
    'a': _parse_a_line,
    'family': _parse_family_line,
    'p': _parse_p_line,
    'p6': _parse_p6_line,
    'pr': _parse_pr_line,
    'id': _parse_id_line,
  }

  @classmethod
  def content(cls, attr = None, exclude = (), sign = False):
    if sign:
      raise NotImplementedError('Signing of %s not implemented' % cls.__name__)

    return _descriptor_content(attr, exclude, (
      ('onion-key', _random_crypto_blob('RSA PUBLIC KEY')),
    ))

  def __init__(self, raw_contents, validate = False, annotations = None):
    super(Microdescriptor, self).__init__(raw_contents, lazy_load = not validate)
    self._annotation_lines = annotations if annotations else []
    entries = _descriptor_components(raw_contents, validate)

    if validate:
      self._parse(entries, validate)
      self._check_constraints(entries)
    else:
      self._entries = entries

  def digest(self, hash_type = DigestHash.SHA256, encoding = DigestEncoding.BASE64):
    """
    Digest of this microdescriptor. These are referenced by...

      * **Microdescriptor Consensus**

        * Referer: :class:`~stem.descriptor.router_status_entry.RouterStatusEntryMicroV3` **digest** attribute
        * Format: **SHA256/BASE64**

    .. versionadded:: 1.8.0

    :param stem.descriptor.DigestHash hash_type: digest hashing algorithm
    :param stem.descriptor.DigestEncoding encoding: digest encoding

    :returns: **hashlib.HASH** or **str** based on our encoding argument
    """

    if hash_type == DigestHash.SHA1:
      return stem.descriptor._encode_digest(hashlib.sha1(self.get_bytes()), encoding)
    elif hash_type == DigestHash.SHA256:
      return stem.descriptor._encode_digest(hashlib.sha256(self.get_bytes()), encoding)
    else:
      raise NotImplementedError('Microdescriptor digests are only available in sha1 and sha256, not %s' % hash_type)

  @lru_cache()
  def get_annotations(self):
    """
    Provides content that appeared prior to the descriptor. If this comes from
    the cached-microdescs then this commonly contains content like...

    ::

      @last-listed 2013-02-24 00:18:30

    :returns: **dict** with the key/value pairs in our annotations
    """

    annotation_dict = {}

    for line in self._annotation_lines:
      if b' ' in line:
        key, value = line.split(b' ', 1)
        annotation_dict[key] = value
      else:
        annotation_dict[line] = None

    return annotation_dict

  def get_annotation_lines(self):
    """
    Provides the lines of content that appeared prior to the descriptor. This
    is the same as the
    :func:`~stem.descriptor.microdescriptor.Microdescriptor.get_annotations`
    results, but with the unparsed lines and ordering retained.

    :returns: **list** with the lines of annotation that came before this descriptor
    """

    return self._annotation_lines

  def _check_constraints(self, entries):
    """
    Does a basic check that the entries conform to this descriptor type's
    constraints.

    :param dict entries: keyword => (value, pgp key) entries

    :raises: **ValueError** if an issue arises in validation
    """

    for keyword in REQUIRED_FIELDS:
      if keyword not in entries:
        raise ValueError("Microdescriptor must have a '%s' entry" % keyword)

    for keyword in SINGLE_FIELDS:
      if keyword in entries and len(entries[keyword]) > 1:
        raise ValueError("The '%s' entry can only appear once in a microdescriptor" % keyword)

    if 'onion-key' != list(entries.keys())[0]:
      raise ValueError("Microdescriptor must start with a 'onion-key' entry")

  def _name(self, is_plural = False):
    return 'microdescriptors' if is_plural else 'microdescriptor'



# Copyright 2014-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Tab completion for our interpreter prompt.
"""

import stem.prereq

from stem.interpreter import uses_settings

if stem.prereq._is_lru_cache_available():
  from functools import lru_cache
else:
  from stem.util.lru_cache import lru_cache


@uses_settings
def _get_commands(controller, config):
  """
  Provides commands recognized by tor.
  """

  commands = config.get('autocomplete', [])

  if controller is None:
    return commands

  # GETINFO commands. Lines are of the form '[option] -- [description]'. This
  # strips '*' from options that accept values.

  results = controller.get_info('info/names', None)

  if results:
    for line in results.splitlines():
      option = line.split(' ', 1)[0].rstrip('*')
      commands.append('GETINFO %s' % option)
  else:
    commands.append('GETINFO ')

  # GETCONF, SETCONF, and RESETCONF commands. Lines are of the form
  # '[option] [type]'.

  results = controller.get_info('config/names', None)

  if results:
    for line in results.splitlines():
      option = line.split(' ', 1)[0]

      commands.append('GETCONF %s' % option)
      commands.append('SETCONF %s' % option)
      commands.append('RESETCONF %s' % option)
  else:
    commands += ['GETCONF ', 'SETCONF ', 'RESETCONF ']

  # SETEVENT, USEFEATURE, and SIGNAL commands. For each of these the GETINFO
  # results are simply a space separated lists of the values they can have.

  options = (
    ('SETEVENTS ', 'events/names'),
    ('USEFEATURE ', 'features/names'),
    ('SIGNAL ', 'signal/names'),
  )

  for prefix, getinfo_cmd in options:
    results = controller.get_info(getinfo_cmd, None)

    if results:
      commands += [prefix + value for value in results.split()]
    else:
      commands.append(prefix)

  # Adds /help commands.

  usage_info = config.get('help.usage', {})

  for cmd in usage_info.keys():
    commands.append('/help ' + cmd)

  return commands


class Autocompleter(object):
  def __init__(self, controller):
    self._commands = _get_commands(controller)

  @lru_cache()
  def matches(self, text):
    """
    Provides autocompletion matches for the given text.

    :param str text: text to check for autocompletion matches with

    :returns: **list** with possible matches
    """

    lowercase_text = text.lower()
    return [cmd for cmd in self._commands if cmd.lower().startswith(lowercase_text)]

  def complete(self, text, state):
    """
    Provides case insensetive autocompletion options, acting as a functor for
    the readlines set_completer function.

    :param str text: text to check for autocompletion matches with
    :param int state: index of result to be provided, readline fetches matches
      until this function provides None

    :returns: **str** with the autocompletion match, **None** if eithe none
      exists or state is higher than our number of matches
    """

    try:
      return self.matches(text)[state]
    except IndexError:
      return None



# Copyright 2012-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

import io
import re

import stem
import stem.control
import stem.descriptor.router_status_entry
import stem.prereq
import stem.response
import stem.util
import stem.version

from stem.util import connection, log, str_tools, tor_tools

# Matches keyword=value arguments. This can't be a simple "(.*)=(.*)" pattern
# because some positional arguments, like circuit paths, can have an equal
# sign.

KW_ARG = re.compile('^(.*) ([A-Za-z0-9_]+)=(\\S*)$')
QUOTED_KW_ARG = re.compile('^(.*) ([A-Za-z0-9_]+)="(.*)"$')
CELL_TYPE = re.compile('^[a-z0-9_]+$')
PARSE_NEWCONSENSUS_EVENTS = True

# TODO: We can remove the following when we drop python2.6 support.

INT_TYPE = int if stem.prereq.is_python_3() else long


class Event(stem.response.ControlMessage):
  """
  Base for events we receive asynchronously, as described in section 4.1 of the
  `control-spec
  <https://gitweb.torproject.org/torspec.git/tree/control-spec.txt>`_.

  :var str type: event type
  :var list positional_args: positional arguments of the event
  :var dict keyword_args: key/value arguments of the event
  """

  _POSITIONAL_ARGS = ()    # attribute names for recognized positional arguments
  _KEYWORD_ARGS = {}       # map of 'keyword => attribute' for recognized attributes
  _QUOTED = ()             # positional arguments that are quoted
  _OPTIONALLY_QUOTED = ()  # positional arguments that may or may not be quoted
  _SKIP_PARSING = False    # skip parsing contents into our positional_args and keyword_args
  _VERSION_ADDED = stem.version.Version('0.1.1.1-alpha')  # minimum version with control-spec V1 event support

  def _parse_message(self):
    if not str(self).strip():
      raise stem.ProtocolError('Received a blank tor event. Events must at the very least have a type.')

    self.type = str(self).split()[0]
    self.positional_args = []
    self.keyword_args = {}

    # if we're a recognized event type then translate ourselves into that subclass

    if self.type in EVENT_TYPE_TO_CLASS:
      self.__class__ = EVENT_TYPE_TO_CLASS[self.type]

    if not self._SKIP_PARSING:
      self._parse_standard_attr()

    self._parse()

  def __hash__(self):
    return stem.util._hash_attr(self, 'arrived_at', parent = stem.response.ControlMessage, cache = True)

  def _parse_standard_attr(self):
    """
    Most events are of the form...
    650 *( positional_args ) *( key "=" value )

    This parses this standard format, populating our **positional_args** and
    **keyword_args** attributes and creating attributes if it's in our event's
    **_POSITIONAL_ARGS** and **_KEYWORD_ARGS**.
    """

    # Tor events contain some number of positional arguments followed by
    # key/value mappings. Parsing keyword arguments from the end until we hit
    # something that isn't a key/value mapping. The rest are positional.

    content = str(self)

    while True:
      match = QUOTED_KW_ARG.match(content)

      if not match:
        match = KW_ARG.match(content)

      if match:
        content, keyword, value = match.groups()
        self.keyword_args[keyword] = value
      else:
        break

    # Setting attributes for the fields that we recognize.

    self.positional_args = content.split()[1:]
    positional = list(self.positional_args)

    for attr_name in self._POSITIONAL_ARGS:
      attr_value = None

      if positional:
        if attr_name in self._QUOTED or (attr_name in self._OPTIONALLY_QUOTED and positional[0].startswith('"')):
          attr_values = [positional.pop(0)]

          if not attr_values[0].startswith('"'):
            raise stem.ProtocolError("The %s value should be quoted, but didn't have a starting quote: %s" % (attr_name, self))

          while True:
            if not positional:
              raise stem.ProtocolError("The %s value should be quoted, but didn't have an ending quote: %s" % (attr_name, self))

            attr_values.append(positional.pop(0))

            if attr_values[-1].endswith('"'):
              break

          attr_value = ' '.join(attr_values)[1:-1]
        else:
          attr_value = positional.pop(0)

      setattr(self, attr_name, attr_value)

    for controller_attr_name, attr_name in self._KEYWORD_ARGS.items():
      setattr(self, attr_name, self.keyword_args.get(controller_attr_name))

  def _iso_timestamp(self, timestamp):
    """
    Parses an iso timestamp (ISOTime2Frac in the control-spec).

    :param str timestamp: timestamp to parse

    :returns: **datetime** with the parsed timestamp

    :raises: :class:`stem.ProtocolError` if timestamp is malformed
    """

    if timestamp is None:
      return None

    try:
      return str_tools._parse_iso_timestamp(timestamp)
    except ValueError as exc:
      raise stem.ProtocolError('Unable to parse timestamp (%s): %s' % (exc, self))

  # method overwritten by our subclasses for special handling that they do
  def _parse(self):
    pass

  def _log_if_unrecognized(self, attr, attr_enum):
    """
    Checks if an attribute exists in a given enumeration, logging a message if
    it isn't. Attributes can either be for a string or collection of strings

    :param str attr: name of the attribute to check
    :param stem.util.enum.Enum enum: enumeration to check against
    """

    attr_values = getattr(self, attr)

    if attr_values:
      if stem.util._is_str(attr_values):
        attr_values = [attr_values]

      for value in attr_values:
        if value not in attr_enum:
          log_id = 'event.%s.unknown_%s.%s' % (self.type.lower(), attr, value)
          unrecognized_msg = "%s event had an unrecognized %s (%s). Maybe a new addition to the control protocol? Full Event: '%s'" % (self.type, attr, value, self)
          log.log_once(log_id, log.INFO, unrecognized_msg)


class AddrMapEvent(Event):
  """
  Event that indicates a new address mapping.

  The ADDRMAP event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha.

  .. versionchanged:: 1.1.0
     Added the cached attribute.

  :var str hostname: address being resolved
  :var str destination: destination of the resolution, this is usually an ip,
    but could be a hostname if TrackHostExits is enabled or **NONE** if the
    resolution failed
  :var datetime expiry: expiration time of the resolution in local time
  :var str error: error code if the resolution failed
  :var datetime utc_expiry: expiration time of the resolution in UTC
  :var bool cached: **True** if the resolution will be kept until it expires,
    **False** otherwise or **None** if undefined
  """

  _POSITIONAL_ARGS = ('hostname', 'destination', 'expiry')
  _KEYWORD_ARGS = {
    'error': 'error',
    'EXPIRES': 'utc_expiry',
    'CACHED': 'cached',
  }
  _OPTIONALLY_QUOTED = ('expiry')

  def _parse(self):
    if self.destination == '<error>':
      self.destination = None

    if self.expiry is not None:
      if self.expiry == 'NEVER':
        self.expiry = None
      else:
        try:
          self.expiry = stem.util.str_tools._parse_timestamp(self.expiry)
        except ValueError:
          raise stem.ProtocolError('Unable to parse date in ADDRMAP event: %s' % self)

    if self.utc_expiry is not None:
      self.utc_expiry = stem.util.str_tools._parse_timestamp(self.utc_expiry)

    if self.cached is not None:
      if self.cached == 'YES':
        self.cached = True
      elif self.cached == 'NO':
        self.cached = False
      else:
        raise stem.ProtocolError("An ADDRMAP event's CACHED mapping can only be 'YES' or 'NO': %s" % self)


class AuthDirNewDescEvent(Event):
  """
  Event specific to directory authorities, indicating that we just received new
  descriptors. The descriptor type contained within this event is unspecified
  so the descriptor contents are left unparsed.

  The AUTHDIR_NEWDESCS event was introduced in tor version 0.1.1.10-alpha and
  removed in 0.3.2.1-alpha. (:spec:`6e887ba`)

  .. deprecated:: 1.6.0
     Tor dropped this event as of version 0.3.2.1. (:spec:`6e887ba`)

  :var stem.AuthDescriptorAction action: what is being done with the descriptor
  :var str message: explanation of why we chose this action
  :var str descriptor: content of the descriptor
  """

  _SKIP_PARSING = True
  _VERSION_ADDED = stem.version.Requirement.EVENT_AUTHDIR_NEWDESCS

  def _parse(self):
    lines = str(self).split('\n')

    if len(lines) < 5:
      raise stem.ProtocolError("AUTHDIR_NEWDESCS events must contain lines for at least the type, action, message, descriptor, and terminating 'OK'")
    elif lines[-1] != 'OK':
      raise stem.ProtocolError("AUTHDIR_NEWDESCS doesn't end with an 'OK'")

    # TODO: For stem 2.0.0 we should consider changing 'descriptor' to a
    # ServerDescriptor instance.

    self.action = lines[1]
    self.message = lines[2]
    self.descriptor = '\n'.join(lines[3:-1])


class BandwidthEvent(Event):
  """
  Event emitted every second with the bytes sent and received by tor.

  The BW event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha.

  :var int read: bytes received by tor that second
  :var int written: bytes sent by tor that second
  """

  _POSITIONAL_ARGS = ('read', 'written')

  def _parse(self):
    if not self.read:
      raise stem.ProtocolError('BW event is missing its read value')
    elif not self.written:
      raise stem.ProtocolError('BW event is missing its written value')
    elif not self.read.isdigit() or not self.written.isdigit():
      raise stem.ProtocolError("A BW event's bytes sent and received should be a positive numeric value, received: %s" % self)

    self.read = INT_TYPE(self.read)
    self.written = INT_TYPE(self.written)


class BuildTimeoutSetEvent(Event):
  """
  Event indicating that the timeout value for a circuit has changed. This was
  first added in tor version 0.2.2.7.

  The BUILDTIMEOUT_SET event was introduced in tor version 0.2.2.7-alpha.

  :var stem.TimeoutSetType set_type: way in which the timeout is changing
  :var int total_times: circuit build times tor used to determine the timeout
  :var int timeout: circuit timeout value in milliseconds
  :var int xm: Pareto parameter Xm in milliseconds
  :var float alpha: Pareto parameter alpha
  :var float quantile: CDF quantile cutoff point
  :var float timeout_rate: ratio of circuits that have time out
  :var int close_timeout: duration to keep measurement circuits in milliseconds
  :var float close_rate: ratio of measurement circuits that are closed
  """

  _POSITIONAL_ARGS = ('set_type',)
  _KEYWORD_ARGS = {
    'TOTAL_TIMES': 'total_times',
    'TIMEOUT_MS': 'timeout',
    'XM': 'xm',
    'ALPHA': 'alpha',
    'CUTOFF_QUANTILE': 'quantile',
    'TIMEOUT_RATE': 'timeout_rate',
    'CLOSE_MS': 'close_timeout',
    'CLOSE_RATE': 'close_rate',
  }
  _VERSION_ADDED = stem.version.Requirement.EVENT_BUILDTIMEOUT_SET

  def _parse(self):
    # convert our integer and float parameters

    for param in ('total_times', 'timeout', 'xm', 'close_timeout'):
      param_value = getattr(self, param)

      if param_value is not None:
        try:
          setattr(self, param, int(param_value))
        except ValueError:
          raise stem.ProtocolError('The %s of a BUILDTIMEOUT_SET should be an integer: %s' % (param, self))

    for param in ('alpha', 'quantile', 'timeout_rate', 'close_rate'):
      param_value = getattr(self, param)

      if param_value is not None:
        try:
          setattr(self, param, float(param_value))
        except ValueError:
          raise stem.ProtocolError('The %s of a BUILDTIMEOUT_SET should be a float: %s' % (param, self))

    self._log_if_unrecognized('set_type', stem.TimeoutSetType)


class CircuitEvent(Event):
  """
  Event that indicates that a circuit has changed.

  The fingerprint or nickname values in our 'path' may be **None** if the
  VERBOSE_NAMES feature isn't enabled. The option was first introduced in tor
  version 0.1.2.2, and on by default after 0.2.2.1.

  The CIRC event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha.

  .. versionchanged:: 1.4.0
     Added the socks_username and socks_password attributes which is used for
     `stream isolation
     <https://gitweb.torproject.org/torspec.git/tree/proposals/171-separate-streams.txt>`_.

  :var str id: circuit identifier
  :var stem.CircStatus status: reported status for the circuit
  :var tuple path: relays involved in the circuit, these are
    **(fingerprint, nickname)** tuples
  :var tuple build_flags: :data:`~stem.CircBuildFlag` attributes
    governing how the circuit is built
  :var stem.CircPurpose purpose: purpose that the circuit is intended for
  :var stem.HiddenServiceState hs_state: status if this is a hidden service circuit
  :var str rend_query: circuit's rendezvous-point if this is hidden service related
  :var datetime created: time when the circuit was created or cannibalized
  :var stem.CircClosureReason reason: reason for the circuit to be closed
  :var stem.CircClosureReason remote_reason: remote side's reason for the circuit to be closed
  :var str socks_username: username for using this circuit
  :var str socks_password: password for using this circuit
  """

  _POSITIONAL_ARGS = ('id', 'status', 'path')
  _KEYWORD_ARGS = {
    'BUILD_FLAGS': 'build_flags',
    'PURPOSE': 'purpose',
    'HS_STATE': 'hs_state',
    'REND_QUERY': 'rend_query',
    'TIME_CREATED': 'created',
    'REASON': 'reason',
    'REMOTE_REASON': 'remote_reason',
    'SOCKS_USERNAME': 'socks_username',
    'SOCKS_PASSWORD': 'socks_password',
  }

  def _parse(self):
    self.path = tuple(stem.control._parse_circ_path(self.path))
    self.created = self._iso_timestamp(self.created)

    if self.build_flags is not None:
      self.build_flags = tuple(self.build_flags.split(','))

    if not tor_tools.is_valid_circuit_id(self.id):
      raise stem.ProtocolError("Circuit IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))

    self._log_if_unrecognized('status', stem.CircStatus)
    self._log_if_unrecognized('build_flags', stem.CircBuildFlag)
    self._log_if_unrecognized('purpose', stem.CircPurpose)
    self._log_if_unrecognized('hs_state', stem.HiddenServiceState)
    self._log_if_unrecognized('reason', stem.CircClosureReason)
    self._log_if_unrecognized('remote_reason', stem.CircClosureReason)

  def _compare(self, other, method):
    # sorting circuit events by their identifier

    if not isinstance(other, CircuitEvent):
      return False

    my_id = getattr(self, 'id')
    their_id = getattr(other, 'id')

    return method(my_id, their_id) if my_id != their_id else method(hash(self), hash(other))

  def __gt__(self, other):
    return self._compare(other, lambda s, o: s > o)

  def __ge__(self, other):
    return self._compare(other, lambda s, o: s >= o)


class CircMinorEvent(Event):
  """
  Event providing information about minor changes in our circuits. This was
  first added in tor version 0.2.3.11.

  The CIRC_MINOR event was introduced in tor version 0.2.3.11-alpha.

  :var str id: circuit identifier
  :var stem.CircEvent event: type of change in the circuit
  :var tuple path: relays involved in the circuit, these are
    **(fingerprint, nickname)** tuples
  :var tuple build_flags: :data:`~stem.CircBuildFlag` attributes
    governing how the circuit is built
  :var stem.CircPurpose purpose: purpose that the circuit is intended for
  :var stem.HiddenServiceState hs_state: status if this is a hidden service circuit
  :var str rend_query: circuit's rendezvous-point if this is hidden service related
  :var datetime created: time when the circuit was created or cannibalized
  :var stem.CircPurpose old_purpose: prior purpose for the circuit
  :var stem.HiddenServiceState old_hs_state: prior status as a hidden service circuit
  """

  _POSITIONAL_ARGS = ('id', 'event', 'path')
  _KEYWORD_ARGS = {
    'BUILD_FLAGS': 'build_flags',
    'PURPOSE': 'purpose',
    'HS_STATE': 'hs_state',
    'REND_QUERY': 'rend_query',
    'TIME_CREATED': 'created',
    'OLD_PURPOSE': 'old_purpose',
    'OLD_HS_STATE': 'old_hs_state',
  }
  _VERSION_ADDED = stem.version.Requirement.EVENT_CIRC_MINOR

  def _parse(self):
    self.path = tuple(stem.control._parse_circ_path(self.path))
    self.created = self._iso_timestamp(self.created)

    if self.build_flags is not None:
      self.build_flags = tuple(self.build_flags.split(','))

    if not tor_tools.is_valid_circuit_id(self.id):
      raise stem.ProtocolError("Circuit IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))

    self._log_if_unrecognized('event', stem.CircEvent)
    self._log_if_unrecognized('build_flags', stem.CircBuildFlag)
    self._log_if_unrecognized('purpose', stem.CircPurpose)
    self._log_if_unrecognized('hs_state', stem.HiddenServiceState)
    self._log_if_unrecognized('old_purpose', stem.CircPurpose)
    self._log_if_unrecognized('old_hs_state', stem.HiddenServiceState)


class ClientsSeenEvent(Event):
  """
  Periodic event on bridge relays that provides a summary of our users.

  The CLIENTS_SEEN event was introduced in tor version 0.2.1.10-alpha.

  :var datetime start_time: time in UTC that we started collecting these stats
  :var dict locales: mapping of country codes to a rounded count for the number of users
  :var dict ip_versions: mapping of ip protocols to a rounded count for the number of users
  """

  _KEYWORD_ARGS = {
    'TimeStarted': 'start_time',
    'CountrySummary': 'locales',
    'IPVersions': 'ip_versions',
  }
  _VERSION_ADDED = stem.version.Requirement.EVENT_CLIENTS_SEEN

  def _parse(self):
    if self.start_time is not None:
      self.start_time = stem.util.str_tools._parse_timestamp(self.start_time)

    if self.locales is not None:
      locale_to_count = {}

      for entry in self.locales.split(','):
        if '=' not in entry:
          raise stem.ProtocolError("The CLIENTS_SEEN's CountrySummary should be a comma separated listing of '<locale>=<count>' mappings: %s" % self)

        locale, count = entry.split('=', 1)

        if len(locale) != 2:
          raise stem.ProtocolError("Locales should be a two character code, got '%s': %s" % (locale, self))
        elif not count.isdigit():
          raise stem.ProtocolError('Locale count was non-numeric (%s): %s' % (count, self))
        elif locale in locale_to_count:
          raise stem.ProtocolError("CountrySummary had multiple mappings for '%s': %s" % (locale, self))

        locale_to_count[locale] = int(count)

      self.locales = locale_to_count

    if self.ip_versions is not None:
      protocol_to_count = {}

      for entry in self.ip_versions.split(','):
        if '=' not in entry:
          raise stem.ProtocolError("The CLIENTS_SEEN's IPVersions should be a comma separated listing of '<protocol>=<count>' mappings: %s" % self)

        protocol, count = entry.split('=', 1)

        if not count.isdigit():
          raise stem.ProtocolError('IP protocol count was non-numeric (%s): %s' % (count, self))

        protocol_to_count[protocol] = int(count)

      self.ip_versions = protocol_to_count


class ConfChangedEvent(Event):
  """
  Event that indicates that our configuration changed, either in response to a
  SETCONF or RELOAD signal.

  The CONF_CHANGED event was introduced in tor version 0.2.3.3-alpha.

  .. deprecated:: 1.7.0
     Deprecated the *config* attribute. Some tor configuration options (like
     ExitPolicy) can have multiple values, so a simple 'str => str' mapping
     meant that we only provided the last.

  .. versionchanged:: 1.7.0
     Added the changed and unset attributes.

  :var dict changed: mapping of configuration options to a list of their new
    values
  :var list unset: configuration options that have been unset
  """

  _SKIP_PARSING = True
  _VERSION_ADDED = stem.version.Requirement.EVENT_CONF_CHANGED

  def _parse(self):
    self.changed = {}
    self.unset = []
    self.config = {}  # TODO: remove in stem 2.0

    # Skip first and last line since they're the header and footer. For
    # instance...
    #
    # 650-CONF_CHANGED
    # 650-ExitNodes=caerSidi
    # 650-ExitPolicy
    # 650-MaxCircuitDirtiness=20
    # 650 OK

    for line in str(self).splitlines()[1:-1]:
      if '=' in line:
        key, value = line.split('=', 1)
        self.changed.setdefault(key, []).append(value)
      else:
        key, value = line, None
        self.unset.append(key)

      self.config[key] = value


class DescChangedEvent(Event):
  """
  Event that indicates that our descriptor has changed.

  The DESCCHANGED event was introduced in tor version 0.1.2.2-alpha.
  """

  _VERSION_ADDED = stem.version.Requirement.EVENT_DESCCHANGED


class GuardEvent(Event):
  """
  Event that indicates that our guard relays have changed. The 'endpoint' could
  be either a...

  * fingerprint
  * 'fingerprint=nickname' pair

  The derived 'endpoint_*' attributes are generally more useful.

  The GUARD event was introduced in tor version 0.1.2.5-alpha.

  :var stem.GuardType guard_type: purpose the guard relay is for
  :var str endpoint: relay that the event concerns
  :var str endpoint_fingerprint: endpoint's finterprint
  :var str endpoint_nickname: endpoint's nickname if it was provided
  :var stem.GuardStatus status: status of the guard relay
  """

  _VERSION_ADDED = stem.version.Requirement.EVENT_GUARD
  _POSITIONAL_ARGS = ('guard_type', 'endpoint', 'status')

  def _parse(self):
    self.endpoint_fingerprint = None
    self.endpoint_nickname = None

    try:
      self.endpoint_fingerprint, self.endpoint_nickname = \
        stem.control._parse_circ_entry(self.endpoint)
    except stem.ProtocolError:
      raise stem.ProtocolError("GUARD's endpoint doesn't match a ServerSpec: %s" % self)

    self._log_if_unrecognized('guard_type', stem.GuardType)
    self._log_if_unrecognized('status', stem.GuardStatus)


class HSDescEvent(Event):
  """
  Event triggered when we fetch a hidden service descriptor that currently isn't in our cache.

  The HS_DESC event was introduced in tor version 0.2.5.2-alpha.

  .. versionadded:: 1.2.0

  .. versionchanged:: 1.3.0
     Added the reason attribute.

  .. versionchanged:: 1.5.0
     Added the replica attribute.

  .. versionchanged:: 1.7.0
     Added the index attribute.

  :var stem.HSDescAction action: what is happening with the descriptor
  :var str address: hidden service address
  :var stem.HSAuth authentication: service's authentication method
  :var str directory: hidden service directory servicing the request
  :var str directory_fingerprint: hidden service directory's finterprint
  :var str directory_nickname: hidden service directory's nickname if it was provided
  :var str descriptor_id: descriptor identifier
  :var stem.HSDescReason reason: reason the descriptor failed to be fetched
  :var int replica: replica number the descriptor involves
  :var str index: computed index of the HSDir the descriptor was uploaded to or fetched from
  """

  _VERSION_ADDED = stem.version.Requirement.EVENT_HS_DESC
  _POSITIONAL_ARGS = ('action', 'address', 'authentication', 'directory', 'descriptor_id')
  _KEYWORD_ARGS = {'REASON': 'reason', 'REPLICA': 'replica', 'HSDIR_INDEX': 'index'}

  def _parse(self):
    self.directory_fingerprint = None
    self.directory_nickname = None

    if self.directory != 'UNKNOWN':
      try:
        self.directory_fingerprint, self.directory_nickname = \
          stem.control._parse_circ_entry(self.directory)
      except stem.ProtocolError:
        raise stem.ProtocolError("HS_DESC's directory doesn't match a ServerSpec: %s" % self)

    if self.replica is not None:
      if not self.replica.isdigit():
        raise stem.ProtocolError('HS_DESC event got a non-numeric replica count (%s): %s' % (self.replica, self))

      self.replica = int(self.replica)

    self._log_if_unrecognized('action', stem.HSDescAction)
    self._log_if_unrecognized('authentication', stem.HSAuth)


class HSDescContentEvent(Event):
  """
  Provides the content of hidden service descriptors we fetch.

  The HS_DESC_CONTENT event was introduced in tor version 0.2.7.1-alpha.

  .. versionadded:: 1.4.0

  :var str address: hidden service address
  :var str descriptor_id: descriptor identifier
  :var str directory: hidden service directory servicing the request
  :var str directory_fingerprint: hidden service directory's finterprint
  :var str directory_nickname: hidden service directory's nickname if it was provided
  :var stem.descriptor.hidden_service.HiddenServiceDescriptorV2 descriptor: descriptor that was retrieved
  """

  _VERSION_ADDED = stem.version.Requirement.EVENT_HS_DESC_CONTENT
  _POSITIONAL_ARGS = ('address', 'descriptor_id', 'directory')

  def _parse(self):
    if self.address == 'UNKNOWN':
      self.address = None

    self.directory_fingerprint = None
    self.directory_nickname = None

    try:
      self.directory_fingerprint, self.directory_nickname = \
        stem.control._parse_circ_entry(self.directory)
    except stem.ProtocolError:
      raise stem.ProtocolError("HS_DESC_CONTENT's directory doesn't match a ServerSpec: %s" % self)

    # skip the first line (our positional arguments) and last ('OK')

    desc_content = str_tools._to_bytes('\n'.join(str(self).splitlines()[1:-1]))
    self.descriptor = None

    if desc_content:
      self.descriptor = list(stem.descriptor.hidden_service._parse_file(io.BytesIO(desc_content)))[0]


class LogEvent(Event):
  """
  Tor logging event. These are the most visible kind of event since, by
  default, tor logs at the NOTICE :data:`~stem.Runlevel` to stdout.

  The logging events were some of the first Control Protocol V1 events
  and were introduced in tor version 0.1.1.1-alpha.

  :var stem.Runlevel runlevel: runlevel of the logged message
  :var str message: logged message
  """

  _SKIP_PARSING = True

  def _parse(self):
    self.runlevel = self.type
    self._log_if_unrecognized('runlevel', stem.Runlevel)

    # message is our content, minus the runlevel and ending "OK" if a
    # multi-line message

    self.message = str(self)[len(self.runlevel) + 1:].rstrip('\nOK')


class NetworkStatusEvent(Event):
  """
  Event for when our copy of the consensus has changed. This was introduced in
  tor version 0.1.2.3.

  The NS event was introduced in tor version 0.1.2.3-alpha.

  :var list desc: :class:`~stem.descriptor.router_status_entry.RouterStatusEntryV3` for the changed descriptors
  """

  _SKIP_PARSING = True
  _VERSION_ADDED = stem.version.Requirement.EVENT_NS

  def _parse(self):
    content = str(self).lstrip('NS\n').rstrip('\nOK')

    # TODO: For stem 2.0.0 consider changing 'desc' to 'descriptors' to match
    # our other events.

    self.desc = list(stem.descriptor.router_status_entry._parse_file(
      io.BytesIO(str_tools._to_bytes(content)),
      False,
      entry_class = stem.descriptor.router_status_entry.RouterStatusEntryV3,
    ))


class NetworkLivenessEvent(Event):
  """
  Event for when the network becomes reachable or unreachable.

  The NETWORK_LIVENESS event was introduced in tor version 0.2.7.2-alpha.

  .. versionadded:: 1.5.0

  :var str status: status of the network ('UP', 'DOWN', or possibly other
    statuses in the future)
  """

  _VERSION_ADDED = stem.version.Requirement.EVENT_NETWORK_LIVENESS
  _POSITIONAL_ARGS = ('status',)


class NewConsensusEvent(Event):
  """
  Event for when we have a new consensus. This is similar to
  :class:`~stem.response.events.NetworkStatusEvent`, except that it contains
  the whole consensus so anything not listed is implicitly no longer
  recommended.

  The NEWCONSENSUS event was introduced in tor version 0.2.1.13-alpha.

  .. versionchanged:: 1.6.0
     Added the consensus_content attribute.

  .. deprecated:: 1.6.0
     In Stem 2.0 we'll remove the desc attribute, so this event only provides
     the unparsed consensus. Callers can then parse it if they'd like. To drop
     parsing before then you can set...

     ::

       stem.response.events.PARSE_NEWCONSENSUS_EVENTS = False

  :var str consensus_content: consensus content
  :var list desc: :class:`~stem.descriptor.router_status_entry.RouterStatusEntryV3` for the changed descriptors
  """

  _SKIP_PARSING = True
  _VERSION_ADDED = stem.version.Requirement.EVENT_NEWCONSENSUS

  def _parse(self):
    self.consensus_content = str(self).lstrip('NEWCONSENSUS\n').rstrip('\nOK')

    # TODO: For stem 2.0.0 consider changing 'desc' to 'descriptors' to match
    # our other events.

    if PARSE_NEWCONSENSUS_EVENTS:
      self.desc = list(stem.descriptor.router_status_entry._parse_file(
        io.BytesIO(str_tools._to_bytes(self.consensus_content)),
        False,
        entry_class = stem.descriptor.router_status_entry.RouterStatusEntryV3,
      ))
    else:
      self.desc = None


class NewDescEvent(Event):
  """
  Event that indicates that a new descriptor is available.

  The fingerprint or nickname values in our 'relays' may be **None** if the
  VERBOSE_NAMES feature isn't enabled. The option was first introduced in tor
  version 0.1.2.2, and on by default after 0.2.2.1.

  The NEWDESC event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha.

  :var tuple relays: **(fingerprint, nickname)** tuples for the relays with
    new descriptors
  """

  def _parse(self):
    self.relays = tuple([stem.control._parse_circ_entry(entry) for entry in str(self).split()[1:]])


class ORConnEvent(Event):
  """
  Event that indicates a change in a relay connection. The 'endpoint' could be
  any of several things including a...

  * fingerprint
  * nickname
  * 'fingerprint=nickname' pair
  * address:port

  The derived 'endpoint_*' attributes are generally more useful.

  The ORCONN event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha. Its id attribute was added in
  version 0.2.5.2-alpha.

  .. versionchanged:: 1.2.0
     Added the id attribute.

  :var str id: connection identifier
  :var str endpoint: relay that the event concerns
  :var str endpoint_fingerprint: endpoint's finterprint if it was provided
  :var str endpoint_nickname: endpoint's nickname if it was provided
  :var str endpoint_address: endpoint's address if it was provided
  :var int endpoint_port: endpoint's port if it was provided
  :var stem.ORStatus status: state of the connection
  :var stem.ORClosureReason reason: reason for the connection to be closed
  :var int circ_count: number of established and pending circuits
  """

  _POSITIONAL_ARGS = ('endpoint', 'status')
  _KEYWORD_ARGS = {
    'REASON': 'reason',
    'NCIRCS': 'circ_count',
    'ID': 'id',
  }

  def _parse(self):
    self.endpoint_fingerprint = None
    self.endpoint_nickname = None
    self.endpoint_address = None
    self.endpoint_port = None

    try:
      self.endpoint_fingerprint, self.endpoint_nickname = \
        stem.control._parse_circ_entry(self.endpoint)
    except stem.ProtocolError:
      if ':' not in self.endpoint:
        raise stem.ProtocolError("ORCONN endpoint is neither a relay nor 'address:port': %s" % self)

      address, port = self.endpoint.rsplit(':', 1)

      if not connection.is_valid_port(port):
        raise stem.ProtocolError("ORCONN's endpoint location's port is invalid: %s" % self)

      self.endpoint_address = address
      self.endpoint_port = int(port)

    if self.circ_count is not None:
      if not self.circ_count.isdigit():
        raise stem.ProtocolError('ORCONN event got a non-numeric circuit count (%s): %s' % (self.circ_count, self))

      self.circ_count = int(self.circ_count)

    if self.id and not tor_tools.is_valid_connection_id(self.id):
      raise stem.ProtocolError("Connection IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))

    self._log_if_unrecognized('status', stem.ORStatus)
    self._log_if_unrecognized('reason', stem.ORClosureReason)


class SignalEvent(Event):
  """
  Event that indicates that tor has received and acted upon a signal being sent
  to the process. As of tor version 0.2.4.6 the only signals conveyed by this
  event are...

  * RELOAD
  * DUMP
  * DEBUG
  * NEWNYM
  * CLEARDNSCACHE

  The SIGNAL event was introduced in tor version 0.2.3.1-alpha.

  :var stem.Signal signal: signal that tor received
  """

  _POSITIONAL_ARGS = ('signal',)
  _VERSION_ADDED = stem.version.Requirement.EVENT_SIGNAL

  def _parse(self):
    # log if we recieved an unrecognized signal
    expected_signals = (
      stem.Signal.RELOAD,
      stem.Signal.DUMP,
      stem.Signal.DEBUG,
      stem.Signal.NEWNYM,
      stem.Signal.CLEARDNSCACHE,
    )

    self._log_if_unrecognized('signal', expected_signals)


class StatusEvent(Event):
  """
  Notification of a change in tor's state. These are generally triggered for
  the same sort of things as log messages of the NOTICE level or higher.
  However, unlike :class:`~stem.response.events.LogEvent` these contain well
  formed data.

  The STATUS_GENERAL, STATUS_CLIENT, STATUS_SERVER events were introduced
  in tor version 0.1.2.3-alpha.

  :var stem.StatusType status_type: category of the status event
  :var stem.Runlevel runlevel: runlevel of the logged message
  :var str action: activity that caused this message
  :var dict arguments: attributes about the event
  """

  _POSITIONAL_ARGS = ('runlevel', 'action')
  _VERSION_ADDED = stem.version.Requirement.EVENT_STATUS

  def _parse(self):
    if self.type == 'STATUS_GENERAL':
      self.status_type = stem.StatusType.GENERAL
    elif self.type == 'STATUS_CLIENT':
      self.status_type = stem.StatusType.CLIENT
    elif self.type == 'STATUS_SERVER':
      self.status_type = stem.StatusType.SERVER
    else:
      raise ValueError("BUG: Unrecognized status type (%s), likely an EVENT_TYPE_TO_CLASS addition without revising how 'status_type' is assigned." % self.type)

    # Just an alias for our parent class' keyword_args since that already
    # parses these for us. Unlike our other event types Tor commonly supplies
    # arbitrary key/value pairs for these, so making an alias here to better
    # draw attention that the StatusEvent will likely have them.

    self.arguments = self.keyword_args

    self._log_if_unrecognized('runlevel', stem.Runlevel)


class StreamEvent(Event):
  """
  Event that indicates that a stream has changed.

  The STREAM event was one of the first Control Protocol V1 events and was
  introduced in tor version 0.1.1.1-alpha.

  :var str id: stream identifier
  :var stem.StreamStatus status: reported status for the stream
  :var str circ_id: circuit that the stream is attached to, this is **None** of
    the stream is unattached
  :var str target: destination of the stream
  :var str target_address: destination address (ip, hostname, or '(Tor_internal)')
  :var int target_port: destination port
  :var stem.StreamClosureReason reason: reason for the stream to be closed
  :var stem.StreamClosureReason remote_reason: remote side's reason for the stream to be closed
  :var stem.StreamSource source: origin of the REMAP request
  :var str source_addr: requester of the connection
  :var str source_address: requester address (ip or hostname)
  :var int source_port: requester port
  :var stem.StreamPurpose purpose: purpose for the stream
  """

  _POSITIONAL_ARGS = ('id', 'status', 'circ_id', 'target')
  _KEYWORD_ARGS = {
    'REASON': 'reason',
    'REMOTE_REASON': 'remote_reason',
    'SOURCE': 'source',
    'SOURCE_ADDR': 'source_addr',
    'PURPOSE': 'purpose',
  }

  def _parse(self):
    if self.target is None:
      raise stem.ProtocolError("STREAM event didn't have a target: %s" % self)
    else:
      if ':' not in self.target:
        raise stem.ProtocolError("Target location must be of the form 'address:port': %s" % self)

      address, port = self.target.rsplit(':', 1)

      if not connection.is_valid_port(port, allow_zero = True):
        raise stem.ProtocolError("Target location's port is invalid: %s" % self)

      self.target_address = address
      self.target_port = int(port)

    if self.source_addr is None:
      self.source_address = None
      self.source_port = None
    else:
      if ':' not in self.source_addr:
        raise stem.ProtocolError("Source location must be of the form 'address:port': %s" % self)

      address, port = self.source_addr.rsplit(':', 1)

      if not connection.is_valid_port(port, allow_zero = True):
        raise stem.ProtocolError("Source location's port is invalid: %s" % self)

      self.source_address = address
      self.source_port = int(port)

    # spec specifies a circ_id of zero if the stream is unattached

    if self.circ_id == '0':
      self.circ_id = None

    self._log_if_unrecognized('reason', stem.StreamClosureReason)
    self._log_if_unrecognized('remote_reason', stem.StreamClosureReason)
    self._log_if_unrecognized('purpose', stem.StreamPurpose)


class StreamBwEvent(Event):
  """
  Event (emitted approximately every second) with the bytes sent and received
  by the application since the last such event on this stream.

  The STREAM_BW event was introduced in tor version 0.1.2.8-beta.

  .. versionchanged:: 1.6.0
     Added the time attribute.

  :var str id: stream identifier
  :var int written: bytes sent by the application
  :var int read: bytes received by the application
  :var datetime time: time when the measurement was recorded
  """

  _POSITIONAL_ARGS = ('id', 'written', 'read', 'time')
  _VERSION_ADDED = stem.version.Requirement.EVENT_STREAM_BW

  def _parse(self):
    if not tor_tools.is_valid_stream_id(self.id):
      raise stem.ProtocolError("Stream IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))
    elif not self.written:
      raise stem.ProtocolError('STREAM_BW event is missing its written value')
    elif not self.read:
      raise stem.ProtocolError('STREAM_BW event is missing its read value')
    elif not self.read.isdigit() or not self.written.isdigit():
      raise stem.ProtocolError("A STREAM_BW event's bytes sent and received should be a positive numeric value, received: %s" % self)

    self.read = INT_TYPE(self.read)
    self.written = INT_TYPE(self.written)
    self.time = self._iso_timestamp(self.time)


class TransportLaunchedEvent(Event):
  """
  Event triggered when a pluggable transport is launched.

  The TRANSPORT_LAUNCHED event was introduced in tor version 0.2.5.0-alpha.

  .. versionadded:: 1.1.0

  :var str type: 'server' or 'client'
  :var str name: name of the pluggable transport
  :var str address: IPv4 or IPv6 address where the transport is listening for
    connections
  :var int port: port where the transport is listening for connections
  """

  _POSITIONAL_ARGS = ('type', 'name', 'address', 'port')
  _VERSION_ADDED = stem.version.Requirement.EVENT_TRANSPORT_LAUNCHED

  def _parse(self):
    if self.type not in ('server', 'client'):
      raise stem.ProtocolError("Transport type should either be 'server' or 'client': %s" % self)

    if not connection.is_valid_ipv4_address(self.address) and \
       not connection.is_valid_ipv6_address(self.address):
      raise stem.ProtocolError("Transport address isn't a valid IPv4 or IPv6 address: %s" % self)

    if not connection.is_valid_port(self.port):
      raise stem.ProtocolError('Transport port is invalid: %s' % self)

    self.port = int(self.port)


class ConnectionBandwidthEvent(Event):
  """
  Event emitted every second with the bytes sent and received by tor on a
  per-connection basis.

  The CONN_BW event was introduced in tor version 0.2.5.2-alpha.

  .. versionadded:: 1.2.0

  .. versionchanged:: 1.6.0
     Renamed 'type' attribute to 'conn_type' so it wouldn't be override parent
     class attribute with the same name.

  :var str id: connection identifier
  :var stem.ConnectionType conn_type: connection type
  :var int read: bytes received by tor that second
  :var int written: bytes sent by tor that second
  """

  _KEYWORD_ARGS = {
    'ID': 'id',
    'TYPE': 'conn_type',
    'READ': 'read',
    'WRITTEN': 'written',
  }

  _VERSION_ADDED = stem.version.Requirement.EVENT_CONN_BW

  def _parse(self):
    if not self.id:
      raise stem.ProtocolError('CONN_BW event is missing its id')
    elif not self.conn_type:
      raise stem.ProtocolError('CONN_BW event is missing its connection type')
    elif not self.read:
      raise stem.ProtocolError('CONN_BW event is missing its read value')
    elif not self.written:
      raise stem.ProtocolError('CONN_BW event is missing its written value')
    elif not self.read.isdigit() or not self.written.isdigit():
      raise stem.ProtocolError("A CONN_BW event's bytes sent and received should be a positive numeric value, received: %s" % self)
    elif not tor_tools.is_valid_connection_id(self.id):
      raise stem.ProtocolError("Connection IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))

    self.read = INT_TYPE(self.read)
    self.written = INT_TYPE(self.written)

    self._log_if_unrecognized('conn_type', stem.ConnectionType)


class CircuitBandwidthEvent(Event):
  """
  Event emitted every second with the bytes sent and received by tor on a
  per-circuit basis.

  The CIRC_BW event was introduced in tor version 0.2.5.2-alpha.

  .. versionadded:: 1.2.0

  .. versionchanged:: 1.6.0
     Added the time attribute.

  .. versionchanged:: 1.7.0
     Added the delivered_read, delivered_written, overhead_read, and
     overhead_written attributes.

  :var str id: circuit identifier
  :var int read: bytes received by tor that second
  :var int written: bytes sent by tor that second
  :var int delivered_read: user payload received by tor that second
  :var int delivered_written: user payload sent by tor that second
  :var int overhead_read: padding so read cells will have a fixed length
  :var int overhead_written: padding so written cells will have a fixed length
  :var datetime time: time when the measurement was recorded
  """

  _KEYWORD_ARGS = {
    'ID': 'id',
    'READ': 'read',
    'WRITTEN': 'written',
    'DELIVERED_READ': 'delivered_read',
    'DELIVERED_WRITTEN': 'delivered_written',
    'OVERHEAD_READ': 'overhead_read',
    'OVERHEAD_WRITTEN': 'overhead_written',
    'TIME': 'time',
  }

  _VERSION_ADDED = stem.version.Requirement.EVENT_CIRC_BW

  def _parse(self):
    if not self.id:
      raise stem.ProtocolError('CIRC_BW event is missing its id')
    elif not self.read:
      raise stem.ProtocolError('CIRC_BW event is missing its read value')
    elif not self.written:
      raise stem.ProtocolError('CIRC_BW event is missing its written value')
    elif not self.read.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's bytes received should be a positive numeric value, received: %s" % self)
    elif not self.written.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's bytes sent should be a positive numeric value, received: %s" % self)
    elif self.delivered_read and not self.delivered_read.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's delivered bytes received should be a positive numeric value, received: %s" % self)
    elif self.delivered_written and not self.delivered_written.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's delivered bytes sent should be a positive numeric value, received: %s" % self)
    elif self.overhead_read and not self.overhead_read.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's overhead bytes received should be a positive numeric value, received: %s" % self)
    elif self.overhead_written and not self.overhead_written.isdigit():
      raise stem.ProtocolError("A CIRC_BW event's overhead bytes sent should be a positive numeric value, received: %s" % self)
    elif not tor_tools.is_valid_circuit_id(self.id):
      raise stem.ProtocolError("Circuit IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))

    self.time = self._iso_timestamp(self.time)

    for attr in ('read', 'written', 'delivered_read', 'delivered_written', 'overhead_read', 'overhead_written'):
      value = getattr(self, attr)

      if value:
        setattr(self, attr, INT_TYPE(value))


class CellStatsEvent(Event):
  """
  Event emitted every second with a count of the number of cells types broken
  down by the circuit. **These events are only emitted if TestingTorNetwork is
  set.**

  The CELL_STATS event was introduced in tor version 0.2.5.2-alpha.

  .. versionadded:: 1.2.0

  :var str id: circuit identifier
  :var str inbound_queue: inbound queue identifier
  :var str inbound_connection: inbound connection identifier
  :var dict inbound_added: mapping of added inbound cell types to their count
  :var dict inbound_removed: mapping of removed inbound cell types to their count
  :var dict inbound_time: mapping of inbound cell types to the time they took to write in milliseconds
  :var str outbound_queue: outbound queue identifier
  :var str outbound_connection: outbound connection identifier
  :var dict outbound_added: mapping of added outbound cell types to their count
  :var dict outbound_removed: mapping of removed outbound cell types to their count
  :var dict outbound_time: mapping of outbound cell types to the time they took to write in milliseconds
  """

  _KEYWORD_ARGS = {
    'ID': 'id',
    'InboundQueue': 'inbound_queue',
    'InboundConn': 'inbound_connection',
    'InboundAdded': 'inbound_added',
    'InboundRemoved': 'inbound_removed',
    'InboundTime': 'inbound_time',
    'OutboundQueue': 'outbound_queue',
    'OutboundConn': 'outbound_connection',
    'OutboundAdded': 'outbound_added',
    'OutboundRemoved': 'outbound_removed',
    'OutboundTime': 'outbound_time',
  }

  _VERSION_ADDED = stem.version.Requirement.EVENT_CELL_STATS

  def _parse(self):
    if self.id and not tor_tools.is_valid_circuit_id(self.id):
      raise stem.ProtocolError("Circuit IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))
    elif self.inbound_queue and not tor_tools.is_valid_circuit_id(self.inbound_queue):
      raise stem.ProtocolError("Queue IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.inbound_queue, self))
    elif self.inbound_connection and not tor_tools.is_valid_connection_id(self.inbound_connection):
      raise stem.ProtocolError("Connection IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.inbound_connection, self))
    elif self.outbound_queue and not tor_tools.is_valid_circuit_id(self.outbound_queue):
      raise stem.ProtocolError("Queue IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.outbound_queue, self))
    elif self.outbound_connection and not tor_tools.is_valid_connection_id(self.outbound_connection):
      raise stem.ProtocolError("Connection IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.outbound_connection, self))

    self.inbound_added = _parse_cell_type_mapping(self.inbound_added)
    self.inbound_removed = _parse_cell_type_mapping(self.inbound_removed)
    self.inbound_time = _parse_cell_type_mapping(self.inbound_time)
    self.outbound_added = _parse_cell_type_mapping(self.outbound_added)
    self.outbound_removed = _parse_cell_type_mapping(self.outbound_removed)
    self.outbound_time = _parse_cell_type_mapping(self.outbound_time)


class TokenBucketEmptyEvent(Event):
  """
  Event emitted when refilling an empty token bucket. **These events are only
  emitted if TestingTorNetwork is set.**

  The TB_EMPTY event was introduced in tor version 0.2.5.2-alpha.

  .. versionadded:: 1.2.0

  :var stem.TokenBucket bucket: bucket being refilled
  :var str id: connection identifier
  :var int read: time in milliseconds since the read bucket was last refilled
  :var int written: time in milliseconds since the write bucket was last refilled
  :var int last_refill: time in milliseconds the bucket has been empty since last refilled
  """

  _POSITIONAL_ARGS = ('bucket',)
  _KEYWORD_ARGS = {
    'ID': 'id',
    'READ': 'read',
    'WRITTEN': 'written',
    'LAST': 'last_refill',
  }

  _VERSION_ADDED = stem.version.Requirement.EVENT_TB_EMPTY

  def _parse(self):
    if self.id and not tor_tools.is_valid_connection_id(self.id):
      raise stem.ProtocolError("Connection IDs must be one to sixteen alphanumeric characters, got '%s': %s" % (self.id, self))
    elif not self.read.isdigit():
      raise stem.ProtocolError("A TB_EMPTY's READ value should be a positive numeric value, received: %s" % self)
    elif not self.written.isdigit():
      raise stem.ProtocolError("A TB_EMPTY's WRITTEN value should be a positive numeric value, received: %s" % self)
    elif not self.last_refill.isdigit():
      raise stem.ProtocolError("A TB_EMPTY's LAST value should be a positive numeric value, received: %s" % self)

    self.read = int(self.read)
    self.written = int(self.written)
    self.last_refill = int(self.last_refill)

    self._log_if_unrecognized('bucket', stem.TokenBucket)


def _parse_cell_type_mapping(mapping):
  """
  Parses a mapping of the form...

    key1:value1,key2:value2...

  ... in which keys are strings and values are integers.

  :param str mapping: value to be parsed

  :returns: dict of **str => int** mappings

  :rasies: **stem.ProtocolError** if unable to parse the mapping
  """

  if mapping is None:
    return None

  results = {}

  for entry in mapping.split(','):
    if ':' not in entry:
      raise stem.ProtocolError("Mappings are expected to be of the form 'key:value', got '%s': %s" % (entry, mapping))

    key, value = entry.rsplit(':', 1)

    if not CELL_TYPE.match(key):
      raise stem.ProtocolError("Key had invalid characters, got '%s': %s" % (key, mapping))
    elif not value.isdigit():
      raise stem.ProtocolError("Values should just be integers, got '%s': %s" % (value, mapping))

    results[key] = int(value)

  return results


EVENT_TYPE_TO_CLASS = {
  'ADDRMAP': AddrMapEvent,
  'AUTHDIR_NEWDESCS': AuthDirNewDescEvent,
  'BUILDTIMEOUT_SET': BuildTimeoutSetEvent,
  'BW': BandwidthEvent,
  'CELL_STATS': CellStatsEvent,
  'CIRC': CircuitEvent,
  'CIRC_BW': CircuitBandwidthEvent,
  'CIRC_MINOR': CircMinorEvent,
  'CLIENTS_SEEN': ClientsSeenEvent,
  'CONF_CHANGED': ConfChangedEvent,
  'CONN_BW': ConnectionBandwidthEvent,
  'DEBUG': LogEvent,
  'DESCCHANGED': DescChangedEvent,
  'ERR': LogEvent,
  'GUARD': GuardEvent,
  'HS_DESC': HSDescEvent,
  'HS_DESC_CONTENT': HSDescContentEvent,
  'INFO': LogEvent,
  'NETWORK_LIVENESS': NetworkLivenessEvent,
  'NEWCONSENSUS': NewConsensusEvent,
  'NEWDESC': NewDescEvent,
  'NOTICE': LogEvent,
  'NS': NetworkStatusEvent,
  'ORCONN': ORConnEvent,
  'SIGNAL': SignalEvent,
  'STATUS_CLIENT': StatusEvent,
  'STATUS_GENERAL': StatusEvent,
  'STATUS_SERVER': StatusEvent,
  'STREAM': StreamEvent,
  'STREAM_BW': StreamBwEvent,
  'TB_EMPTY': TokenBucketEmptyEvent,
  'TRANSPORT_LAUNCHED': TransportLaunchedEvent,
  'WARN': LogEvent,

  # accounting for a bug in tor 0.2.0.22
  'STATUS_SEVER': StatusEvent,
}



# Copyright 2012-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Connection and networking based utility functions.

**Module Overview:**

::

  download - download from a given url
  get_connections - quieries the connections belonging to a given process
  system_resolvers - provides connection resolution methods that are likely to be available
  port_usage - brief description of the common usage for a port

  is_valid_ipv4_address - checks if a string is a valid IPv4 address
  is_valid_ipv6_address - checks if a string is a valid IPv6 address
  is_valid_port - checks if something is a valid representation for a port
  is_private_address - checks if an IPv4 address belongs to a private range or not

  address_to_int - provides an integer representation of an IP address

  expand_ipv6_address - provides an IPv6 address with its collapsed portions expanded
  get_mask_ipv4 - provides the mask representation for a given number of bits
  get_mask_ipv6 - provides the IPv6 mask representation for a given number of bits

.. data:: Resolver (enum)

  Method for resolving a process' connections.

  .. versionadded:: 1.1.0

  .. versionchanged:: 1.4.0
     Added **NETSTAT_WINDOWS**.

  .. versionchanged:: 1.6.0
     Added **BSD_FSTAT**.

  .. deprecated:: 1.6.0
     The SOCKSTAT connection resolver is proving to be unreliable
     (:trac:`23057`), and will be dropped in the 2.0.0 release unless fixed.

  ====================  ===========
  Resolver              Description
  ====================  ===========
  **PROC**              /proc contents
  **NETSTAT**           netstat
  **NETSTAT_WINDOWS**   netstat command under Windows
  **SS**                ss command
  **LSOF**              lsof command
  **SOCKSTAT**          sockstat command under \\*nix
  **BSD_SOCKSTAT**      sockstat command under FreeBSD
  **BSD_PROCSTAT**      procstat command under FreeBSD
  **BSD_FSTAT**         fstat command under OpenBSD
  ====================  ===========
"""

import collections
import os
import platform
import re
import socket
import sys
import time

import stem
import stem.util
import stem.util.proc
import stem.util.system

from stem.util import conf, enum, log, str_tools

try:
  # account for urllib's change between python 2.x and 3.x
  import urllib.request as urllib
except ImportError:
  import urllib2 as urllib

# Connection resolution is risky to log about since it's highly likely to
# contain sensitive information. That said, it's also difficult to get right in
# a platform independent fashion. To opt into the logging requried to
# troubleshoot connection resolution set the following...

LOG_CONNECTION_RESOLUTION = False

Resolver = enum.Enum(
  ('PROC', 'proc'),
  ('NETSTAT', 'netstat'),
  ('NETSTAT_WINDOWS', 'netstat (windows)'),
  ('SS', 'ss'),
  ('LSOF', 'lsof'),
  ('SOCKSTAT', 'sockstat'),
  ('BSD_SOCKSTAT', 'sockstat (bsd)'),
  ('BSD_PROCSTAT', 'procstat (bsd)'),
  ('BSD_FSTAT', 'fstat (bsd)')
)

FULL_IPv4_MASK = '255.255.255.255'
FULL_IPv6_MASK = 'FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF'

PORT_USES = None  # port number => description

RESOLVER_COMMAND = {
  Resolver.PROC: '',

  # -n = prevents dns lookups, -p = include process, -W = don't crop addresses (needed for ipv6)
  Resolver.NETSTAT: 'netstat -npW',

  # -a = show all TCP/UDP connections, -n = numeric addresses and ports, -o = include pid
  Resolver.NETSTAT_WINDOWS: 'netstat -ano',

  # -n = numeric ports, -p = include process, -t = tcp sockets, -u = udp sockets
  Resolver.SS: 'ss -nptu',

  # -n = prevent dns lookups, -P = show port numbers (not names), -i = ip only, -w = no warnings
  # (lsof provides a '-p <pid>' but oddly in practice it seems to be ~11-28% slower)
  Resolver.LSOF: 'lsof -wnPi',

  Resolver.SOCKSTAT: 'sockstat',

  # -4 = IPv4, -c = connected sockets
  Resolver.BSD_SOCKSTAT: 'sockstat -4c',

  # -f <pid> = process pid
  Resolver.BSD_PROCSTAT: 'procstat -f {pid}',

  # -p <pid> = process pid
  Resolver.BSD_FSTAT: 'fstat -p {pid}',
}

RESOLVER_FILTER = {
  Resolver.PROC: '',

  # tcp        0    586 192.168.0.1:44284       38.229.79.2:443         ESTABLISHED 15843/tor
  Resolver.NETSTAT: '^{protocol}\\s+.*\\s+{local}\\s+{remote}\\s+ESTABLISHED\\s+{pid}/{name}\\s*$',

  # tcp        586 192.168.0.1:44284       38.229.79.2:443         ESTABLISHED 15843
  Resolver.NETSTAT_WINDOWS: '^\\s*{protocol}\\s+{local}\\s+{remote}\\s+ESTABLISHED\\s+{pid}\\s*$',

  # tcp    ESTAB      0      0           192.168.0.20:44415       38.229.79.2:443    users:(("tor",15843,9))
  Resolver.SS: '^{protocol}\\s+ESTAB\\s+.*\\s+{local}\\s+{remote}\\s+users:\\(\\("{name}",(?:pid=)?{pid},(?:fd=)?[0-9]+\\)\\)$',

  # tor  3873  atagar  45u  IPv4  40994  0t0  TCP 10.243.55.20:45724->194.154.227.109:9001 (ESTABLISHED)
  Resolver.LSOF: '^{name}\\s+{pid}\\s+.*\\s+{protocol}\\s+{local}->{remote} \\(ESTABLISHED\\)$',

  # atagar   tor                  15843    tcp4   192.168.0.20:44092        68.169.35.102:443         ESTABLISHED
  Resolver.SOCKSTAT: '^\\S+\\s+{name}\\s+{pid}\\s+{protocol}4\\s+{local}\\s+{remote}\\s+ESTABLISHED$',

  # _tor     tor        4397  12 tcp4   172.27.72.202:54011   127.0.0.1:9001
  Resolver.BSD_SOCKSTAT: '^\\S+\\s+{name}\\s+{pid}\\s+\\S+\\s+{protocol}4\\s+{local}\\s+{remote}$',

  # 3561 tor                 4 s - rw---n--   2       0 TCP 10.0.0.2:9050 10.0.0.1:22370
  Resolver.BSD_PROCSTAT: '^\\s*{pid}\\s+{name}\\s+.*\\s+{protocol}\\s+{local}\\s+{remote}$',

  # _tor     tor        15843   20* internet stream tcp 0x0 192.168.1.100:36174 --> 4.3.2.1:443
  Resolver.BSD_FSTAT: '^\\S+\\s+{name}\\s+{pid}\\s+.*\\s+{protocol}\\s+\\S+\\s+{local}\\s+[-<]-[->]\\s+{remote}$',
}


class Connection(collections.namedtuple('Connection', ['local_address', 'local_port', 'remote_address', 'remote_port', 'protocol', 'is_ipv6'])):
  """
  Network connection information.

  .. versionchanged:: 1.5.0
     Added the **is_ipv6** attribute.

  :var str local_address: ip address the connection originates from
  :var int local_port: port the connection originates from
  :var str remote_address: destionation ip address
  :var int remote_port: destination port
  :var str protocol: protocol of the connection ('tcp', 'udp', etc)
  :var bool is_ipv6: addresses are ipv6 if true, and ipv4 otherwise
  """


def download(url, timeout = None, retries = None):
  """
  Download from the given url.

  .. versionadded:: 1.8.0

  :param str url: uncompressed url to download from
  :param int timeout: timeout when connection becomes idle, no timeout applied
    if **None**
  :param int retires: maximum attempts to impose

  :returns: **bytes** content of the given url

  :raises:
    * :class:`~stem.DownloadTimeout` if our request timed out
    * :class:`~stem.DownloadFailed` if our request fails
  """

  if retries is None:
    retries = 0

  start_time = time.time()

  try:
    return urllib.urlopen(url, timeout = timeout).read()
  except socket.timeout as exc:
    raise stem.DownloadTimeout(url, exc, sys.exc_info()[2], timeout)
  except:
    exc, stacktrace = sys.exc_info()[1:3]

    if timeout is not None:
      timeout -= time.time() - start_time

    if retries > 0 and (timeout is None or timeout > 0):
      log.debug('Failed to download from %s (%i retries remaining): %s' % (url, retries, exc))
      return download(url, timeout, retries - 1)
    else:
      log.debug('Failed to download from %s: %s' % (url, exc))
      raise stem.DownloadFailed(url, exc, stacktrace)


def get_connections(resolver = None, process_pid = None, process_name = None):
  """
  Retrieves a list of the current connections for a given process. This
  provides a list of :class:`~stem.util.connection.Connection`. Note that
  addresses may be IPv4 *or* IPv6 depending on what the platform supports.

  .. versionadded:: 1.1.0

  .. versionchanged:: 1.5.0
     Made our resolver argument optional.

  .. versionchanged:: 1.5.0
     IPv6 support when resolving via proc, netstat, lsof, or ss.

  :param Resolver resolver: method of connection resolution to use, if not
    provided then one is picked from among those that should likely be
    available for the system
  :param int process_pid: pid of the process to retrieve
  :param str process_name: name of the process to retrieve

  :returns: **list** of :class:`~stem.util.connection.Connection` instances

  :raises:
    * **ValueError** if neither a process_pid nor process_name is provided

    * **IOError** if no connections are available or resolution fails
      (generally they're indistinguishable). The common causes are the
      command being unavailable or permissions.
  """

  if not resolver:
    available_resolvers = system_resolvers()

    if available_resolvers:
      resolver = available_resolvers[0]
    else:
      raise IOError('Unable to determine a connection resolver')

  if not process_pid and not process_name:
    raise ValueError('You must provide a pid or process name to provide connections for')

  def _log(msg):
    if LOG_CONNECTION_RESOLUTION:
      log.debug(msg)

  _log('=' * 80)
  _log('Querying connections for resolver: %s, pid: %s, name: %s' % (resolver, process_pid, process_name))

  if isinstance(process_pid, str):
    try:
      process_pid = int(process_pid)
    except ValueError:
      raise ValueError('Process pid was non-numeric: %s' % process_pid)

  if process_pid is None:
    all_pids = stem.util.system.pid_by_name(process_name, True)

    if len(all_pids) == 0:
      if resolver in (Resolver.NETSTAT_WINDOWS, Resolver.PROC, Resolver.BSD_PROCSTAT):
        raise IOError("Unable to determine the pid of '%s'. %s requires the pid to provide the connections." % (process_name, resolver))
    elif len(all_pids) == 1:
      process_pid = all_pids[0]
    else:
      if resolver in (Resolver.NETSTAT_WINDOWS, Resolver.PROC, Resolver.BSD_PROCSTAT):
        raise IOError("There's multiple processes named '%s'. %s requires a single pid to provide the connections." % (process_name, resolver))

  if resolver == Resolver.PROC:
    return stem.util.proc.connections(pid = process_pid)

  resolver_command = RESOLVER_COMMAND[resolver].format(pid = process_pid)

  try:
    results = stem.util.system.call(resolver_command)
  except OSError as exc:
    raise IOError("Unable to query '%s': %s" % (resolver_command, exc))

  resolver_regex_str = RESOLVER_FILTER[resolver].format(
    protocol = '(?P<protocol>\\S+)',
    local = '(?P<local>[\\[\\]0-9a-f.:]+)',
    remote = '(?P<remote>[\\[\\]0-9a-f.:]+)',
    pid = process_pid if process_pid else '[0-9]*',
    name = process_name if process_name else '\\S*',
  )

  _log('Resolver regex: %s' % resolver_regex_str)
  _log('Resolver results:\n%s' % '\n'.join(results))

  connections = []
  resolver_regex = re.compile(resolver_regex_str)

  def _parse_address_str(addr_type, addr_str, line):
    addr, port = addr_str.rsplit(':', 1)

    if not is_valid_ipv4_address(addr) and not is_valid_ipv6_address(addr, allow_brackets = True):
      _log('Invalid %s address (%s): %s' % (addr_type, addr, line))
      return None, None
    elif not is_valid_port(port):
      _log('Invalid %s port (%s): %s' % (addr_type, port, line))
      return None, None
    else:
      _log('Valid %s:%s: %s' % (addr, port, line))
      return addr.lstrip('[').rstrip(']'), int(port)

  for line in results:
    match = resolver_regex.match(line)

    if match:
      attr = match.groupdict()

      local_addr, local_port = _parse_address_str('local', attr['local'], line)
      remote_addr, remote_port = _parse_address_str('remote', attr['remote'], line)

      if not (local_addr and local_port and remote_addr and remote_port):
        continue  # missing or malformed field

      protocol = attr['protocol'].lower()

      if protocol == 'tcp6':
        protocol = 'tcp'

      if protocol not in ('tcp', 'udp'):
        _log('Unrecognized protocol (%s): %s' % (protocol, line))
        continue

      conn = Connection(local_addr, local_port, remote_addr, remote_port, protocol, is_valid_ipv6_address(local_addr))
      connections.append(conn)
      _log(str(conn))

  _log('%i connections found' % len(connections))

  if not connections:
    raise IOError('No results found using: %s' % resolver_command)

  return connections


def system_resolvers(system = None):
  """
  Provides the types of connection resolvers likely to be available on this platform.

  .. versionadded:: 1.1.0

  .. versionchanged:: 1.3.0
     Renamed from get_system_resolvers() to system_resolvers(). The old name
     still works as an alias, but will be dropped in Stem version 2.0.0.

  :param str system: system to get resolvers for, this is determined by
    platform.system() if not provided

  :returns: **list** of :data:`~stem.util.connection.Resolver` instances available on this platform
  """

  if system is None:
    if stem.util.system.is_gentoo():
      system = 'Gentoo'
    else:
      system = platform.system()

  if system == 'Windows':
    resolvers = [Resolver.NETSTAT_WINDOWS]
  elif system == 'Darwin':
    resolvers = [Resolver.LSOF]
  elif system == 'OpenBSD':
    resolvers = [Resolver.BSD_FSTAT]
  elif system == 'FreeBSD':
    # Netstat is available, but lacks a '-p' equivalent so we can't associate
    # the results to processes. The platform also has a ss command, but it
    # belongs to a spreadsheet application.

    resolvers = [Resolver.BSD_SOCKSTAT, Resolver.BSD_PROCSTAT, Resolver.LSOF]
  else:
    # Sockstat isn't available by default on ubuntu.

    resolvers = [Resolver.NETSTAT, Resolver.SOCKSTAT, Resolver.LSOF, Resolver.SS]

  # remove any that aren't in the user's PATH

  resolvers = [r for r in resolvers if stem.util.system.is_available(RESOLVER_COMMAND[r])]

  # proc resolution, by far, outperforms the others so defaults to this is able

  if stem.util.proc.is_available() and os.access('/proc/net/tcp', os.R_OK) and os.access('/proc/net/udp', os.R_OK):
    resolvers = [Resolver.PROC] + resolvers

  return resolvers


def port_usage(port):
  """
  Provides the common use of a given port. For example, 'HTTP' for port 80 or
  'SSH' for 22.

  .. versionadded:: 1.2.0

  :param int port: port number to look up

  :returns: **str** with a description for the port, **None** if none is known
  """

  global PORT_USES

  if PORT_USES is None:
    config = conf.Config()
    config_path = os.path.join(os.path.dirname(__file__), 'ports.cfg')

    try:
      config.load(config_path)
      port_uses = {}

      for key, value in config.get('port', {}).items():
        if key.isdigit():
         port_uses[int(key)] = value
        elif '-' in key:
          min_port, max_port = key.split('-', 1)

          for port_entry in range(int(min_port), int(max_port) + 1):
            port_uses[port_entry] = value
        else:
          raise ValueError("'%s' is an invalid key" % key)

      PORT_USES = port_uses
    except Exception as exc:
      log.warn("BUG: stem failed to load its internal port descriptions from '%s': %s" % (config_path, exc))

  if not PORT_USES:
    return None

  if isinstance(port, str) and port.isdigit():
    port = int(port)

  return PORT_USES.get(port)


def is_valid_ipv4_address(address):
  """
  Checks if a string is a valid IPv4 address.

  :param str address: string to be checked

  :returns: **True** if input is a valid IPv4 address, **False** otherwise
  """

  if isinstance(address, bytes):
    address = str_tools._to_unicode(address)
  elif not stem.util._is_str(address):
    return False

  # checks if theres four period separated values

  if address.count('.') != 3:
     return False

  # checks that each value in the octet are decimal values between 0-255
  for entry in address.split('.'):
    if not entry.isdigit() or int(entry) < 0 or int(entry) > 255:
      return False
    elif entry[0] == '0' and len(entry) > 1:
      return False  # leading zeros, for instance in '1.2.3.001'

  return True


def is_valid_ipv6_address(address, allow_brackets = False):
  """
  Checks if a string is a valid IPv6 address.

  :param str address: string to be checked
  :param bool allow_brackets: ignore brackets which form '[address]'

  :returns: **True** if input is a valid IPv6 address, **False** otherwise
  """

  if isinstance(address, bytes):
    address = str_tools._to_unicode(address)
  elif not stem.util._is_str(address):
    return False

  if allow_brackets:
    if address.startswith('[') and address.endswith(']'):
      address = address[1:-1]

  if address.count('.') == 3:
    # Likely an ipv4-mapped portion. Check that its vaild, then replace with a
    # filler.

    ipv4_start = address.rfind(':', 0, address.find('.')) + 1
    ipv4_end = address.find(':', ipv4_start + 1)

    if ipv4_end == -1:
      ipv4_end = None  # don't crop the last character

    if not is_valid_ipv4_address(address[ipv4_start:ipv4_end]):
      return False

    addr_comp = [address[:ipv4_start - 1] if ipv4_start != 0 else None, 'ff:ff', address[ipv4_end + 1:] if ipv4_end else None]
    address = ':'.join(filter(None, addr_comp))

  # addresses are made up of eight colon separated groups of four hex digits
  # with leading zeros being optional
  # https://en.wikipedia.org/wiki/IPv6#Address_format

  colon_count = address.count(':')

  if colon_count > 7:
    return False  # too many groups
  elif colon_count != 7 and '::' not in address:
    return False  # not enough groups and none are collapsed
  elif address.count('::') > 1 or ':::' in address:
    return False  # multiple groupings of zeros can't be collapsed

  for entry in address.split(':'):
    if not re.match('^[0-9a-fA-f]{0,4}$', entry):
      return False

  return True


def is_valid_port(entry, allow_zero = False):
  """
  Checks if a string or int is a valid port number.

  :param list,str,int entry: string, integer or list to be checked
  :param bool allow_zero: accept port number of zero (reserved by definition)

  :returns: **True** if input is an integer and within the valid port range, **False** otherwise
  """

  try:
    value = int(entry)

    if str(value) != str(entry):
      return False  # invalid leading char, e.g. space or zero
    elif allow_zero and value == 0:
      return True
    else:
      return value > 0 and value < 65536
  except TypeError:
    if isinstance(entry, (tuple, list)):
      for port in entry:
        if not is_valid_port(port, allow_zero):
          return False

      return True
    else:
      return False
  except ValueError:
    return False


def is_private_address(address):
  """
  Checks if the IPv4 address is in a range belonging to the local network or
  loopback. These include:

    * Private ranges: 10.*, 172.16.* - 172.31.*, 192.168.*
    * Loopback: 127.*

  .. versionadded:: 1.1.0

  :param str address: string to be checked

  :returns: **True** if input is in a private range, **False** otherwise

  :raises: **ValueError** if the address isn't a valid IPv4 address
  """

  if not is_valid_ipv4_address(address):
    raise ValueError("'%s' isn't a valid IPv4 address" % address)

  # checks for any of the simple wildcard ranges

  if address.startswith('10.') or address.startswith('192.168.') or address.startswith('127.'):
    return True

  # checks for the 172.16.* - 172.31.* range

  if address.startswith('172.'):
    second_octet = int(address.split('.')[1])

    if second_octet >= 16 and second_octet <= 31:
      return True

  return False


def address_to_int(address):
  """
  Provides an integer representation of a IPv4 or IPv6 address that can be used
  for sorting.

  .. versionadded:: 1.5.0

  :param str address: IPv4 or IPv6 address

  :returns: **int** representation of the address
  """

  # TODO: Could be neat to also use this for serialization if we also had an
  # int_to_address() function.

  return int(_address_to_binary(address), 2)


def expand_ipv6_address(address):
  """
  Expands abbreviated IPv6 addresses to their full colon separated hex format.
  For instance...

  ::

    >>> expand_ipv6_address('2001:db8::ff00:42:8329')
    '2001:0db8:0000:0000:0000:ff00:0042:8329'

    >>> expand_ipv6_address('::')
    '0000:0000:0000:0000:0000:0000:0000:0000'

    >>> expand_ipv6_address('::ffff:5.9.158.75')
    '0000:0000:0000:0000:0000:ffff:0509:9e4b'

  :param str address: IPv6 address to be expanded

  :raises: **ValueError** if the address can't be expanded due to being malformed
  """

  if not is_valid_ipv6_address(address):
    raise ValueError("'%s' isn't a valid IPv6 address" % address)

  # expand ipv4-mapped portions of addresses
  if address.count('.') == 3:
    ipv4_start = address.rfind(':', 0, address.find('.')) + 1
    ipv4_end = address.find(':', ipv4_start + 1)

    if ipv4_end == -1:
      ipv4_end = None  # don't crop the last character

    # Converts ipv4 address to its hex ipv6 representation. For instance...
    #
    #   '5.9.158.75' => '0509:9e4b'

    ipv4_bin = _address_to_binary(address[ipv4_start:ipv4_end])
    groupings = [ipv4_bin[16 * i:16 * (i + 1)] for i in range(2)]
    ipv6_snippet = ':'.join(['%04x' % int(group, 2) for group in groupings])

    addr_comp = [address[:ipv4_start - 1] if ipv4_start != 0 else None, ipv6_snippet, address[ipv4_end + 1:] if ipv4_end else None]
    address = ':'.join(filter(None, addr_comp))

  # expands collapsed groupings, there can only be a single '::' in a valid
  # address
  if '::' in address:
    missing_groups = 7 - address.count(':')
    address = address.replace('::', '::' + ':' * missing_groups)

  # inserts missing zeros
  for index in range(8):
    start = index * 5
    end = address.index(':', start) if index != 7 else len(address)
    missing_zeros = 4 - (end - start)

    if missing_zeros > 0:
      address = address[:start] + '0' * missing_zeros + address[start:]

  return address


def get_mask_ipv4(bits):
  """
  Provides the IPv4 mask for a given number of bits, in the dotted-quad format.

  :param int bits: number of bits to be converted

  :returns: **str** with the subnet mask representation for this many bits

  :raises: **ValueError** if given a number of bits outside the range of 0-32
  """

  if bits > 32 or bits < 0:
    raise ValueError('A mask can only be 0-32 bits, got %i' % bits)
  elif bits == 32:
    return FULL_IPv4_MASK

  # get the binary representation of the mask
  mask_bin = _get_binary(2 ** bits - 1, 32)[::-1]

  # breaks it into eight character groupings
  octets = [mask_bin[8 * i:8 * (i + 1)] for i in range(4)]

  # converts each octet into its integer value
  return '.'.join([str(int(octet, 2)) for octet in octets])


def get_mask_ipv6(bits):
  """
  Provides the IPv6 mask for a given number of bits, in the hex colon-delimited
  format.

  :param int bits: number of bits to be converted

  :returns: **str** with the subnet mask representation for this many bits

  :raises: **ValueError** if given a number of bits outside the range of 0-128
  """

  if bits > 128 or bits < 0:
    raise ValueError('A mask can only be 0-128 bits, got %i' % bits)
  elif bits == 128:
    return FULL_IPv6_MASK

  # get the binary representation of the mask
  mask_bin = _get_binary(2 ** bits - 1, 128)[::-1]

  # breaks it into sixteen character groupings
  groupings = [mask_bin[16 * i:16 * (i + 1)] for i in range(8)]

  # converts each group into its hex value
  return ':'.join(['%04x' % int(group, 2) for group in groupings]).upper()


def _get_masked_bits(mask):
  """
  Provides the number of bits that an IPv4 subnet mask represents. Note that
  not all masks can be represented by a bit count.

  :param str mask: mask to be converted

  :returns: **int** with the number of bits represented by the mask

  :raises: **ValueError** if the mask is invalid or can't be converted
  """

  if not is_valid_ipv4_address(mask):
    raise ValueError("'%s' is an invalid subnet mask" % mask)

  # converts octets to binary representation
  mask_bin = _address_to_binary(mask)
  mask_match = re.match('^(1*)(0*)$', mask_bin)

  if mask_match:
    return 32 - len(mask_match.groups()[1])
  else:
    raise ValueError('Unable to convert mask to a bit count: %s' % mask)


def _get_binary(value, bits):
  """
  Provides the given value as a binary string, padded with zeros to the given
  number of bits.

  :param int value: value to be converted
  :param int bits: number of bits to pad to
  """

  # http://www.daniweb.com/code/snippet216539.html
  return ''.join([str((value >> y) & 1) for y in range(bits - 1, -1, -1)])


# TODO: In stem 2.x we should consider unifying this with
# stem.client.datatype's _unpack_ipv4_address() and _unpack_ipv6_address().

def _address_to_binary(address):
  """
  Provides the binary value for an IPv4 or IPv6 address.

  :returns: **str** with the binary representation of this address

  :raises: **ValueError** if address is neither an IPv4 nor IPv6 address
  """

  if is_valid_ipv4_address(address):
    return ''.join([_get_binary(int(octet), 8) for octet in address.split('.')])
  elif is_valid_ipv6_address(address):
    address = expand_ipv6_address(address)
    return ''.join([_get_binary(int(grouping, 16), 16) for grouping in address.split(':')])
  else:
    raise ValueError("'%s' is neither an IPv4 or IPv6 address" % address)


# TODO: drop with stem 2.x
# We renamed our methods to drop a redundant 'get_*' prefix, so alias the old
# names for backward compatability.

get_system_resolvers = system_resolvers



# Copyright 2012-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

import stem.response
import stem.socket


class MapAddressResponse(stem.response.ControlMessage):
  """
  Reply for a MAPADDRESS query.
  Doesn't raise an exception unless no addresses were mapped successfully.

  :var dict entries: mapping between the original and replacement addresses

  :raises:
    * :class:`stem.OperationFailed` if Tor was unable to satisfy the request
    * :class:`stem.InvalidRequest` if the addresses provided were invalid
  """

  def _parse_message(self):
    # Example:
    # 250-127.192.10.10=torproject.org
    # 250 1.2.3.4=tor.freehaven.net

    if not self.is_ok():
      for code, _, message in self.content():
        if code == '512':
          raise stem.InvalidRequest(code, message)
        elif code == '451':
          raise stem.OperationFailed(code, message)
        else:
          raise stem.ProtocolError('MAPADDRESS returned unexpected response code: %s', code)

    self.entries = {}

    for code, _, message in self.content():
      if code == '250':
        try:
          key, value = message.split('=', 1)
          self.entries[key] = value
        except ValueError:
          raise stem.ProtocolError(None, "MAPADDRESS returned '%s', which isn't a mapping" % message)



# Copyright 2015-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Commandline argument parsing for our interpreter prompt.
"""

import collections
import getopt
import os

import stem.interpreter
import stem.util.connection

DEFAULT_ARGS = {
  'control_address': '127.0.0.1',
  'control_port': 'default',
  'user_provided_port': False,
  'control_socket': '/var/run/tor/control',
  'user_provided_socket': False,
  'tor_path': 'tor',
  'run_cmd': None,
  'run_path': None,
  'disable_color': False,
  'print_help': False,
}

OPT = 'i:s:h'
OPT_EXPANDED = ['interface=', 'socket=', 'tor=', 'run=', 'no-color', 'help']


def parse(argv):
  """
  Parses our arguments, providing a named tuple with their values.

  :param list argv: input arguments to be parsed

  :returns: a **named tuple** with our parsed arguments

  :raises: **ValueError** if we got an invalid argument
  """

  args = dict(DEFAULT_ARGS)

  try:
    recognized_args, unrecognized_args = getopt.getopt(argv, OPT, OPT_EXPANDED)

    if unrecognized_args:
      error_msg = "aren't recognized arguments" if len(unrecognized_args) > 1 else "isn't a recognized argument"
      raise getopt.GetoptError("'%s' %s" % ("', '".join(unrecognized_args), error_msg))
  except Exception as exc:
    raise ValueError('%s (for usage provide --help)' % exc)

  for opt, arg in recognized_args:
    if opt in ('-i', '--interface'):
      if ':' in arg:
        address, port = arg.rsplit(':', 1)
      else:
        address, port = None, arg

      if address is not None:
        if not stem.util.connection.is_valid_ipv4_address(address):
          raise ValueError("'%s' isn't a valid IPv4 address" % address)

        args['control_address'] = address

      if not stem.util.connection.is_valid_port(port):
        raise ValueError("'%s' isn't a valid port number" % port)

      args['control_port'] = int(port)
      args['user_provided_port'] = True
    elif opt in ('-s', '--socket'):
      args['control_socket'] = arg
      args['user_provided_socket'] = True
    elif opt in ('--tor'):
      args['tor_path'] = arg
    elif opt in ('--run'):
      if os.path.exists(arg):
        args['run_path'] = arg
      else:
        args['run_cmd'] = arg
    elif opt == '--no-color':
      args['disable_color'] = True
    elif opt in ('-h', '--help'):
      args['print_help'] = True

  # translates our args dict into a named tuple

  Args = collections.namedtuple('Args', args.keys())
  return Args(**args)


def get_help():
  """
  Provides our --help usage information.

  :returns: **str** with our usage information
  """

  return stem.interpreter.msg(
    'msg.help',
    address = DEFAULT_ARGS['control_address'],
    port = DEFAULT_ARGS['control_port'],
    socket = DEFAULT_ARGS['control_socket'],
  )
